"""The Choi channel."""
import numpy as np
from toqito.states import max_entangled


def choi(a_var: int = 1, b_var: int = 1, c_var: int = 0) -> np.ndarray:
    r"""
    Produce the Choi channel or one of its generalizations [Choi92]_.

    The *Choi channel* is a positive map on 3-by-3 matrices that is capable of detecting some
    entanglement that the transpose map is not.

    The standard Choi channel defined with :code:`a=1`, :code:`b=1`, and :code:`c=0` is the Choi
    matrix of the positive map defined in [Choi92]_. Many of these maps are capable of detecting
    PPT entanglement.

    Examples
    ==========

    The standard Choi channel is given as

    .. math::
        \Phi_{1, 1, 0} =
        \begin{pmatrix}
            1 & 0 & 0 & 0 & -1 & 0 & 0 & 0 & -1 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
            -1 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & -1 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
            -1 & 0 & 0 & 0 & -1 & 0 & 0 & 0 & 1
        \end{pmatrix}

    We can generate the Choi channel in :code:`toqito` as follows.

    >>> from toqito.channels import choi
    >>> import numpy as np
    >>> choi()
    [[ 1.,  0.,  0.,  0., -1.,  0.,  0.,  0., -1.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.],
     [-1.,  0.,  0.,  0.,  1.,  0.,  0.,  0., -1.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  0.],
     [-1.,  0.,  0.,  0., -1.,  0.,  0.,  0.,  1.]])

    The reduction channel is the map :math:`R` defined by:

    .. math::
        R(X) = \text{Tr}(X) \mathbb{I} - X.

    The matrix correspond to this is given as

    .. math::
        \Phi_{0, 1, 1} =
        \begin{pmatrix}
            0 & 0 & 0 & 0 & -1 & 0 & 0 & 0 & -1 \\
            0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
            -1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & -1 \\
            0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
            -1 & 0 & 0 & 0 & -1 & 0 & 0 & 0 & 0
        \end{pmatrix}

    The reduction channel is the Choi channel that arises when :code:`a = 0` and when :code:`b =
    c = 1`. We can obtain this matrix using :code:`toqito` as follows.

    >>> from toqito.channels import choi
    >>> import numpy as np
    >>> choi(0, 1, 1)
    [[ 0.,  0.,  0.,  0., -1.,  0.,  0.,  0., -1.],
     [ 0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.],
     [-1.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
     [ 0.,  0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  1.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  0.],
     [-1.,  0.,  0.,  0., -1.,  0.,  0.,  0.,  0.]])

    See Also
    ==========
    reduction

    References
    ==========
    .. [Choi92] Cho, Sung Je, Seung-Hyeok Kye, and Sa Ge Lee.
        "Generalized Choi maps in three-dimensional matrix algebra."
        Linear algebra and its applications 171 (1992): 213-224.
        https://www.sciencedirect.com/science/article/pii/002437959290260H

    :param a_var: Default integer for standard Choi map.
    :param b_var: Default integer for standard Choi map.
    :param c_var: Default integer for standard Choi map.
    :return: The Choi channel (or one of its  generalizations).
    """
    psi = max_entangled(3, False, False)
    return (
        np.diag([a_var + 1, c_var, b_var, b_var, a_var + 1, c_var, c_var, b_var, a_var + 1])
        - psi * psi.conj().T
    )



"""Test is_symmetric."""
import numpy as np

from toqito.matrix_props import is_symmetric


def test_is_symmetric():
    """Test that symmetric matrix returns True."""
    mat = np.array([[1, 7, 3], [7, 4, -5], [3, -5, 6]])
    np.testing.assert_equal(is_symmetric(mat), True)


def test_is_not_symmetric():
    """Test that non-symmetric matrix returns False."""
    mat = np.array([[1, 2], [3, 4]])
    np.testing.assert_equal(is_symmetric(mat), False)


def test_is_symmetric_not_square():
    """Input must be a square matrix."""
    mat = np.array([[-1, 1, 1], [1, 2, 3]])
    np.testing.assert_equal(is_symmetric(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Horodecki state."""
from typing import List

import numpy as np


def horodecki(a_param: float, dim: List[int] = None) -> np.ndarray:
    r"""
    Produce a Horodecki state [HOR]_, [CHR]_.

    Returns the Horodecki state in either :math:`(3 \otimes 3)`-dimensional space or
    :math:`(2 \otimes 4)`-dimensional space, depending on the dimensions in the 1-by-2 vector
    :code:`dim`.

    The Horodecki state was introduced in [1] which serves as an example in
    :math:`\mathbb{C}^3 \otimes \mathbb{C}` or :math:`\mathbb{C}^2 \otimes \mathbb{C}^4` of an
    entangled state that is positive under partial transpose (PPT). The state is PPT for all
    :math:`a \in [0, 1]` and separable only for :code:`a_param = 0` or :code:`a_param = 1`.

    These states have the following definitions:

    .. math::
        \begin{equation}
            \rho_a^{3 \otimes 3} = \frac{1}{8a + 1}
            \begin{pmatrix}
                a & 0 & 0 & 0 & a & 0 & 0 & 0 & a \\
                0 & a & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & a & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & a & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & a & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & a & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & \frac{1}{2}
                \left( 1 + a \right) & 0 & \frac{1}{2} \sqrt{1 - a^2} \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & a & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & \frac{1}{2} \sqrt{1 - a^2} & 0
                & \frac{1}{2} \left(1 + a \right) \\
            \end{pmatrix},
        \end{equation}

    .. math::
        \begin{equation}
            \rho_a^{2 \otimes 4} = \frac{1}{7a + 1}
            \begin{pmatrix}
                a & 0 & 0 & 0 & 0 & a & 0 & 0  \\
                0 & a & 0 & 0 & 0 & 0 & a & 0 \\
                0 & 0 & a & 0 & 0 & 0 & 0 & a  \\
                0 & 0 & 0 & a & 0 & 0 & 0 & 0  \\
                0 & 0 & 0 & 0 & \frac{1}{2} \left(1 + a\right) & 0 & 0
                & \frac{1}{2}\sqrt{1 -a^2} \\
                a & 0 & 0 & 0 & 0 & a & 0 & 0 \\
                0 & a & 0 & 0 & 0 & 0 & a & 0 \\
                0 & 0 & a & 0 & \frac{1}{2}\sqrt{1 - a^2} & 0 & 0
                & \frac{1}{2}\left(1 +a \right)
            \end{pmatrix}.
        \end{equation}

    .. note::
        Refer to [CHR]_ (specifically equations (1) and (2)) for more information on this state
        and its properties. The 3x3 Horodecki state is defined explicitly in Section 4.1 of [HOR]_
        and the 2x4 Horodecki state is defined explicitly in Section 4.2 of [HOR]_.

    Examples
    ==========

    The following code generates a Horodecki state in
    :math:`\mathbb{C}^3 \otimes \mathbb{C}^3`

    >>> from toqito.states import horodecki
    >>> horodecki(0.5, [3, 3])
    [[0.1       , 0.        , 0.        , 0.        , 0.1       ,
      0.        , 0.        , 0.        , 0.1       ],
     [0.        , 0.1       , 0.        , 0.        , 0.        ,
      0.        , 0.        , 0.        , 0.        ],
     [0.        , 0.        , 0.1       , 0.        , 0.        ,
      0.        , 0.        , 0.        , 0.        ],
     [0.        , 0.        , 0.        , 0.1       , 0.        ,
      0.        , 0.        , 0.        , 0.        ],
     [0.1       , 0.        , 0.        , 0.        , 0.1       ,
      0.        , 0.        , 0.        , 0.1       ],
     [0.        , 0.        , 0.        , 0.        , 0.        ,
      0.1       , 0.        , 0.        , 0.        ],
     [0.        , 0.        , 0.        , 0.        , 0.        ,
      0.        , 0.15      , 0.        , 0.08660254],
     [0.        , 0.        , 0.        , 0.        , 0.        ,
      0.        , 0.        , 0.1       , 0.        ],
     [0.1       , 0.        , 0.        , 0.        , 0.1       ,
      0.        , 0.08660254, 0.        , 0.15      ]]

    The following code generates a Horodecki state in :math:`\mathbb{C}^2 \otimes \mathbb{C}^4`.

    >>> from toqito.states import horodecki
    >>> horodecki(0.5, [2, 4])
    [[0.11111111, 0.        , 0.        , 0.        , 0.        ,
      0.11111111, 0.        , 0.        ],
     [0.        , 0.11111111, 0.        , 0.        , 0.        ,
      0.        , 0.11111111, 0.        ],
     [0.        , 0.        , 0.11111111, 0.        , 0.        ,
      0.        , 0.        , 0.11111111],
     [0.        , 0.        , 0.        , 0.11111111, 0.        ,
      0.        , 0.        , 0.        ],
     [0.        , 0.        , 0.        , 0.        , 0.16666667,
      0.        , 0.        , 0.09622504],
     [0.11111111, 0.        , 0.        , 0.        , 0.        ,
      0.11111111, 0.        , 0.        ],
     [0.        , 0.11111111, 0.        , 0.        , 0.        ,
      0.        , 0.11111111, 0.        ],
     [0.        , 0.        , 0.11111111, 0.        , 0.09622504,
      0.        , 0.        , 0.16666667]]

    References
    ==========
    .. [HOR] P. Horodecki.
        Separability criterion and inseparable mixed states with positive
        partial transpose.
        arXiv: 970.3004.

    .. [CHR] K. Chruscinski.
        On the symmetry of the seminal Horodecki state.
        arXiv: 1009.4385.
    """
    if a_param < 0 or a_param > 1:
        raise ValueError("Invalid: Argument A_PARAM must be in the interval " "[0, 1].")

    if dim is None:
        dim = np.array([3, 3])

    if np.array_equal(dim, np.array([3, 3])):
        n_a_param = 1 / (8 * a_param + 1)
        b_param = (1 + a_param) / 2
        c_param = np.sqrt(1 - a_param ** 2) / 2

        horo_state = n_a_param * np.array(
            [
                [a_param, 0, 0, 0, a_param, 0, 0, 0, a_param],
                [0, a_param, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, a_param, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, a_param, 0, 0, 0, 0, 0],
                [a_param, 0, 0, 0, a_param, 0, 0, 0, a_param],
                [0, 0, 0, 0, 0, a_param, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, b_param, 0, c_param],
                [0, 0, 0, 0, 0, 0, 0, a_param, 0],
                [a_param, 0, 0, 0, a_param, 0, c_param, 0, b_param],
            ]
        )
        return horo_state

    if np.array_equal(dim, np.array([2, 4])):
        n_a_param = 1 / (7 * a_param + 1)
        b_param = (1 + a_param) / 2
        c_param = np.sqrt(1 - a_param ** 2) / 2

        horo_state = n_a_param * np.array(
            [
                [a_param, 0, 0, 0, 0, a_param, 0, 0],
                [0, a_param, 0, 0, 0, 0, a_param, 0],
                [0, 0, a_param, 0, 0, 0, 0, a_param],
                [0, 0, 0, a_param, 0, 0, 0, 0],
                [0, 0, 0, 0, b_param, 0, 0, c_param],
                [a_param, 0, 0, 0, 0, a_param, 0, 0],
                [0, a_param, 0, 0, 0, 0, a_param, 0],
                [0, 0, a_param, 0, c_param, 0, 0, b_param],
            ]
        )
        return horo_state
    raise ValueError("InvalidDim: DIM must be one of [3, 3], or [2, 4].")



"""Properties of matrices and vectors."""
from toqito.matrix_props.is_square import is_square
from toqito.matrix_props.is_diagonal import is_diagonal
from toqito.matrix_props.is_normal import is_normal
from toqito.matrix_props.is_symmetric import is_symmetric
from toqito.matrix_props.is_hermitian import is_hermitian
from toqito.matrix_props.is_identity import is_identity
from toqito.matrix_props.is_idempotent import is_idempotent
from toqito.matrix_props.is_positive_semidefinite import is_positive_semidefinite
from toqito.matrix_props.is_positive_definite import is_positive_definite
from toqito.matrix_props.is_density import is_density
from toqito.matrix_props.is_commuting import is_commuting
from toqito.matrix_props.is_projection import is_projection
from toqito.matrix_props.is_unitary import is_unitary
from toqito.matrix_props.majorizes import majorizes



"""Test expr_as_np_array."""
import cvxpy
import numpy as np

from toqito.helper import expr_as_np_array


def test_expr_as_np_array():
    """Ensure return type is numpy object."""
    expr = cvxpy.bmat([[1, 2], [3, 4]])

    res_mat = expr_as_np_array(expr)
    np.testing.assert_equal(isinstance(res_mat, np.ndarray), True)


def test_expr_as_np_array_scalar():
    """Ensure return type is numpy object for scalar expression."""
    cvx_var = cvxpy.Variable()
    np.testing.assert_equal(isinstance(expr_as_np_array(cvx_var), np.ndarray), True)


def test_expr_as_np_array_vector():
    """Ensure return type is numpy object for vector expression."""
    cvx_var = cvxpy.Parameter(5)
    np.testing.assert_equal(isinstance(expr_as_np_array(cvx_var), np.ndarray), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Check if states form mutually unbiased basis."""
from typing import Any, List, Union

import numpy as np


def is_mutually_unbiased_basis(vec_list: List[Union[np.ndarray, List[Union[float, Any]]]]) -> bool:
    r"""
    Check if list of vectors constitute a mutually unbiased basis [WikMUB]_.

    We say that two orthonormal bases

    .. math::
        \begin{equation}
            \mathcal{B}_0 = \left\{u_a: a \in \Sigma \right\} \subset \mathbb{C}^{\Sigma}
            \quad \text{and} \quad
            \mathcal{B}_1 = \left\{v_a: a \in \Sigma \right\} \subset \mathbb{C}^{\Sigma}
        \end{equation}

    are *mutually unbiased* if and only if
    :math:`\left|\langle u_a, v_b \rangle\right| = 1/\sqrt{\Sigma}` for all :math:`a, b \in \Sigma`.

    For :math:`n \in \mathbb{N}`, a set of orthonormal bases :math:`\left\{
    \mathcal{B}_0, \ldots, \mathcal{B}_{n-1} \right\}` are mutually unbiased bases if and only if
    every basis is mutually unbiased with every other basis in the set, i.e. :math:`\mathcal{B}_x`
    is mutually unbiased with :math:`\mathcal{B}_x^{\prime}` for all :math:`x \not= x^{\prime}` with
    :math:`x, x^{\prime} \in \Sigma`.

    Examples
    ==========

    MUB of dimension :math:`2`.

    For :math:`d=2`, the following constitutes a mutually unbiased basis:

    .. math::
        \begin{equation}
            \begin{aligned}
                M_0 &= \left\{ |0 \rangle, |1 \rangle \right\}, \\
                M_1 &= \left\{ \frac{|0 \rangle + |1 \rangle}{\sqrt{2}},
                \frac{|0 \rangle - |1 \rangle}{\sqrt{2}} \right\}, \\
                M_2 &= \left\{ \frac{|0 \rangle i|1 \rangle}{\sqrt{2}},
                \frac{|0 \rangle - i|1 \rangle}{\sqrt{2}} \right\}. \\
            \end{aligned}
        \end{equation}

    >>> import numpy as np
    >>> from toqito.states import basis
    >>> from toqito.state_props import is_mutually_unbiased_basis
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> mub_1 = [e_0, e_1]
    >>> mub_2 = [1 / np.sqrt(2) * (e_0 + e_1), 1 / np.sqrt(2) * (e_0 - e_1)]
    >>> mub_3 = [1 / np.sqrt(2) * (e_0 + 1j * e_1), 1 / np.sqrt(2) * (e_0 - 1j * e_1)]
    >>> mubs = [mub_1, mub_2, mub_3]
    >>> is_mutually_unbiased_basis(mubs)
    True

    Non non-MUB of dimension :math:`2`.

    >>> import numpy as np
    >>> from toqito.states import basis
    >>> from toqito.state_props import is_mutually_unbiased_basis
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> mub_1 = [e_0, e_1]
    >>> mub_2 = [1 / np.sqrt(2) * (e_0 + e_1), e_1]
    >>> mub_3 = [1 / np.sqrt(2) * (e_0 + 1j * e_1), e_0]
    >>> mubs = [mub_1, mub_2, mub_3]
    >>> is_mutually_unbiased_basis(mubs)
    False

    References
    ==========
    .. [WikMUB] Wikipedia: Mutually unbiased bases
        https://en.wikipedia.org/wiki/Mutually_unbiased_bases

    :param vec_list: The list of vectors to check.
    :return: True if :code:`vec_list` constitutes a mutually unbiased basis, and False otherwise.
    """
    if len(vec_list) <= 1:
        raise ValueError("There must be at least two bases provided as input.")

    dim = vec_list[0][0].shape[0]
    for i, _ in enumerate(vec_list):
        for j, _ in enumerate(vec_list):
            for k in range(dim):
                if i != j:
                    if not np.isclose(
                        np.abs(np.inner(vec_list[i][k].conj().T[0], vec_list[j][k].conj().T[0]))
                        ** 2,
                        1 / dim,
                    ):
                        return False
    return True



"""Test partial_trace."""
import cvxpy
import numpy as np

from cvxpy.atoms.affine.vstack import Vstack
from toqito.channels import partial_trace


def test_partial_trace():
    """
    Standard call to partial_trace.

    By default, the partial_trace function takes the trace over the second
    subsystem.
    """
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = np.array([[7, 11], [23, 27]])

    res = partial_trace(test_input_mat)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_sys():
    """
    Specify the `sys` argument.

    By specifying the `sys` argument, you can perform the partial trace
    the first subsystem instead:
    """
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = np.array([[12, 14], [20, 22]])

    res = partial_trace(test_input_mat, 1)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_sys_int_dim_int():
    """
    Default second subsystem.

    By default, the partial_transpose function takes the trace over
    the second subsystem.
    """
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = np.array([[7, 11], [23, 27]])

    res = partial_trace(test_input_mat, 2, 2)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_sys_int_dim_int_2():
    """
    Default second subsystem.

    By default, the partial_transpose function takes the trace over
    the second subsystem.
    """
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = 34

    res = partial_trace(test_input_mat, 2, 1)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_8_by_8():
    """Test for 8-by-8 matrix."""
    test_input_mat = np.arange(1, 65).reshape(8, 8)
    res = partial_trace(test_input_mat, [1, 2], [2, 2, 2])

    expected_res = np.array([[112, 116], [144, 148]])

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_8_by_8_2():
    """Test for 8-by-8 matrix."""
    test_input_mat = np.arange(1, 65).reshape(8, 8)
    res = partial_trace(test_input_mat, [1], [2, 2, 2])

    expected_res = np.array(
        [[38, 40, 42, 44], [54, 56, 58, 60], [70, 72, 74, 76], [86, 88, 90, 92]]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_8_by_8_3():
    """Test for 8-by-8 matrix."""
    test_input_mat = np.arange(1, 65).reshape(8, 8)
    res = partial_trace(test_input_mat, 3, [2, 2, 2])

    expected_res = np.array(
        [[11, 15, 19, 23], [43, 47, 51, 55], [75, 79, 83, 87], [107, 111, 115, 119]]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_16_by_16():
    """Test for 16-by-16 matrix."""
    test_input_mat = np.arange(1, 257).reshape(16, 16)
    res = partial_trace(test_input_mat, [1, 3], [2, 2, 2, 2])

    expected_res = np.array(
        [
            [344, 348, 360, 364],
            [408, 412, 424, 428],
            [600, 604, 616, 620],
            [664, 668, 680, 684],
        ]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_16_by_16_2():
    """Test for 16-by-16 matrix."""
    test_input_mat = np.arange(1, 257).reshape(16, 16)
    res = partial_trace(test_input_mat, [1, 2], [2, 2, 2, 2])

    expected_res = np.array(
        [
            [412, 416, 420, 424],
            [476, 480, 484, 488],
            [540, 544, 548, 552],
            [604, 608, 612, 616],
        ]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_64_by_64():
    """Test for 64-by-64 matrix."""
    test_input_mat = np.arange(1, 4097).reshape(64, 64)
    res = partial_trace(test_input_mat, [1, 2, 3], [4, 4, 4])

    expected_res = np.array([[131104]])

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_64_by_64_2():
    """Test for 64-by-64 matrix."""
    test_input_mat = np.arange(1, 4097).reshape(64, 64)
    res = partial_trace(test_input_mat, [1, 2], [4, 4, 4])

    expected_res = np.array(
        [
            [31216, 31232, 31248, 31264],
            [32240, 32256, 32272, 32288],
            [33264, 33280, 33296, 33312],
            [34288, 34304, 34320, 34336],
        ]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_trace_invalid_sys_arg():
    """The `sys` argument must be either a list or int."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
        partial_trace(rho, "invalid_input")


def test_partial_trace_non_square_matrix_dim_2():
    """Matrix must be square for partial trace."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
        partial_trace(rho, 2, [2])


def test_partial_trace_cvxpy():
    """Test partial trace on cvxpy objects."""
    x_var = cvxpy.Variable((4, 4), hermitian=True)
    x_pt = partial_trace(x_var)
    np.testing.assert_equal(isinstance(x_pt, Vstack), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Symmetric projection operator."""
from itertools import permutations
from scipy import linalg, sparse

import numpy as np

from toqito.perms import permutation_operator


def symmetric_projection(
    dim: int, p_val: int = 2, partial: bool = False
) -> [np.ndarray, sparse.lil_matrix]:
    r"""
    Produce the projection onto the symmetric subspace [CJKLZ14]_.

    For a complex Euclidean space :math:`\mathcal{X}` and a positive integer :math:`n`, the
    projection onto the symmetric subspace is given by

    .. math::
        \frac{1}{n!} \sum_{\pi \in S_n} W_{\pi}

    where :math:`W_{\pi}` is the swap operator and where :math:`S_n` is the symmetric group on
    :math:`n` symbols.

    Produces the orthogonal projection onto the symmetric subspace of :code:`p_val` copies of
    `dim`-dimensional space. If `partial = True`, then the symmetric projection (PS) isn't the
    orthogonal projection itself, but rather a matrix whose columns form an orthonormal basis for
    the symmetric subspace (and hence the PS * PS' is the orthogonal projection onto the symmetric
    subspace).

    This function was adapted from the QETLAB package.

    Examples
    ==========

    The :math:`2`-dimensional symmetric projection with :math:`p=1` is given as
    :math:`2`-by-:math:`2` identity matrix

    .. math::
        \begin{pmatrix}
            1 & 0 \\
            0 & 1
        \end{pmatrix}.

    Using :code:`toqito`, we can see this gives the proper result.

    >>> from toqito.perms import symmetric_projection
    >>> symmetric_projection(2, 1).todense()
    [[1., 0.],
     [0., 1.]]

    When :math:`d = 2` and :math:`p = 2` we have that

    .. math::
        \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1/2 & 1/2 & 0 \\
            0 & 1/2 & 1/2 & 0 \\
            0 & 0 & 0 & 1
        \end{pmatrix}.

    Using :code:`toqito` we can see this gives the proper result.

    >>> from toqito.perms import symmetric_projection
    >>> symmetric_projection(dim=2).todense()
    [[1. , 0. , 0. , 0. ],
     [0. , 0.5, 0.5, 0. ],
     [0. , 0.5, 0.5, 0. ],
     [0. , 0. , 0. , 1. ]]

    References
    ==========
     .. [CJKLZ14] J. Chen, Z. Ji, D. Kribs, N. Lütkenhaus, and B. Zeng.
        "Symmetric extension of two-qubit states".
        Physical Review A 90.3 (2014): 032318.
        https://arxiv.org/abs/1310.3530
        E-print: arXiv:1310.3530 [quant-ph]

    :param dim: The dimension of the local systems.
    :param p_val: Default value of 2.
    :param partial: Default value of 0.
    :return: Projection onto the symmetric subspace.
    """
    dimp = dim ** p_val

    if p_val == 1:
        return np.eye(dim)

    p_list = np.array(list(permutations(np.arange(1, p_val + 1))))
    p_fac = np.math.factorial(p_val)
    sym_proj = np.zeros((dimp, dimp))

    for j in range(p_fac):
        sym_proj += permutation_operator(dim * np.ones(p_val), p_list[j, :], False, True)
    sym_proj = sym_proj / p_fac

    if partial:
        sym_proj = linalg.orth(sym_proj)
    return sym_proj



"""Test max_entangled."""
import numpy as np

from toqito.states import basis
from toqito.states import max_entangled


def test_max_ent_2():
    """Generate maximally entangled state: `1/sqrt(2) * (|00> + |11>)`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = 1 / np.sqrt(2) * (np.kron(e_0, e_0) + np.kron(e_1, e_1))
    res = max_entangled(2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_max_ent_2_0_0():
    """Generate maximally entangled state: `|00> + |11>`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = 1 * (np.kron(e_0, e_0) + np.kron(e_1, e_1))
    res = max_entangled(2, False, False)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Determine if one vector or matrix majorizes another."""
from typing import Union, List
import numpy as np


def majorizes(a_var: Union[np.ndarray, List[int]], b_var: Union[np.ndarray, List[int]]):
    r"""
    Determine if one vector or matrix majorizes another. [WikMajorization]_.

    Given :math:`a, b \in \mathbb{R}^d`, we say that :math:`a` **weakly majorizes** (or dominates)
    :math:`b` from below if and only if

    .. math::
        \sum_{i=1}^k a_i^{\downarrow} \geq \sum_{i=1}^k b_i^{\downarrow}

    for all :math:`k \in \{1, \ldots, d\}`.

    This function was adapted from the QETLAB package.

    Examples
    ==========

    Simple example illustrating that the vector :math:`(3, 0, 0)` majorizes the vector
    :math:`(1, 1, 1)`.

    >>> from toqito.matrix_props import majorizes
    >>> majorizes([3, 0, 0], [1, 1, 1])
    True

    The majorization criterion says that every separable state
    :math:`\rho \in \text{D}(\mathcal{A} \otimes \mathcal{B})` is such that
    :math:`\text{Tr}_{\mathcal{B}}(\rho)` majorizes
    :math:`\text{Tr}_{\mathcal{A}}(\rho)`.

    >>> from toqito.matrix_props import majorizes
    >>> from toqito.states import max_entangled
    >>> from toqito.channels import partial_trace
    >>>
    >>> v_vec = max_entangled(3)
    >>> rho = v_vec * v_vec.conj().T
    >>> majorizes(partial_trace(rho), rho)
    False

    References
    ==========
    .. [WikMajorization] Wikipedia: Majorization
       https://en.wikipedia.org/wiki/Majorization

    :param a_var: Matrix or vector provided as list or np.array.
    :param b_var: Matrix or vector provided as list or np.array.
    :return: Return :code:`True` if :code:`a_var` majorizes :code:`b_var` and :code:`False`
             otherwise.
    """
    # If input if provided as list, convert to np.array.
    if isinstance(a_var, list):
        a_var = np.array(a_var)
    if isinstance(b_var, list):
        b_var = np.array(b_var)

    # If matrix, obtain singular values for majorization.
    if len(a_var.shape) == 1:
        a_var = np.sort(a_var)[::-1]
    # Otherwise, just sort in descending order.
    else:
        _, a_var, _ = np.linalg.svd(a_var)

    # Do the same for second input argument.
    if len(b_var.shape) == 1:
        b_var = np.sort(b_var)[::-1]
    else:
        _, b_var, _ = np.linalg.svd(b_var)

    la_var = len(a_var)
    lb_var = len(b_var)

    # If different length vectors, pad with zeros.
    if la_var < lb_var:
        a_var = np.pad(a_var, (0, lb_var - la_var), "constant")
    elif lb_var < la_var:
        b_var = np.pad(b_var, (0, la_var - lb_var), "constant")

    cta = 0
    ctb = -np.linalg.norm(a_var) * np.finfo(float).eps ** (3 / 4)

    # Check for majorization.
    for k, _ in enumerate(a_var):
        cta = cta + a_var[k]
        ctb = ctb + b_var[k]
        if cta < ctb:
            return False
    return True



"""A number of properties of quantum channels."""
from toqito.channel_props.is_herm_preserving import is_herm_preserving
from toqito.channel_props.is_completely_positive import is_completely_positive
from toqito.channel_props.is_positive import is_positive
from toqito.channel_props.is_unital import is_unital
from toqito.channel_props.choi_rank import choi_rank
from toqito.channel_props.is_trace_preserving import is_trace_preserving



"""Test depolarizing."""
import numpy as np

from toqito.channels import depolarizing
from toqito.channel_ops import apply_channel


def test_depolarizing_complete_depolarizing():
    """Maps every density matrix to the maximally-mixed state."""
    test_input_mat = np.array(
        [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
    )

    expected_res = (
        1 / 4 * np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
    )

    res = apply_channel(test_input_mat, depolarizing(4))

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_depolarizing_partially_depolarizing():
    """The partially depolarizing channel for `p = 0.5`."""
    test_input_mat = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])

    expected_res = np.array(
        [
            [17.125, 0.25, 0.375, 0.5],
            [0.625, 17.75, 0.875, 1],
            [1.125, 1.25, 18.375, 1.5],
            [1.625, 1.75, 1.875, 19],
        ]
    )

    res = apply_channel(test_input_mat, depolarizing(4, 0.5))

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test np_array_as_expr."""
import cvxpy
import numpy as np

from toqito.helper import np_array_as_expr


def test_np_array_as_expr():
    """Ensure return type is CVX object."""
    test_input_mat = np.array([[1, 4, 7], [2, 5, 8], [3, 6, 9]])

    res_mat = np_array_as_expr(test_input_mat)
    np.testing.assert_equal(isinstance(res_mat, cvxpy.atoms.affine.vstack.Vstack), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_povm."""
import numpy as np

from toqito.measurement_props import is_povm
from toqito.random import random_povm


def test_is_povm_true():
    """Test if valid measurement returns True."""
    dim, num_inputs, num_outputs = 2, 2, 2
    measurements = random_povm(dim, num_inputs, num_outputs)

    np.testing.assert_equal(is_povm([measurements[:, :, 0, 0], measurements[:, :, 0, 1]]), True)


def test_is_povm_false():
    """Test if invalid POVM returns False."""
    non_meas_1 = np.array([[1, 2], [3, 4]])
    non_meas_2 = np.array([[5, 6], [7, 8]])
    non_meas = [non_meas_1, non_meas_2]

    np.testing.assert_equal(is_povm(non_meas), False)


def test_is_povm_false_not_sum_identity():
    """Test if invalid POVM (does not sum to identity)."""
    non_meas_1 = np.array([[1, 0], [0, 1]])
    non_meas_2 = np.array([[1, 0], [0, 1]])
    non_meas = [non_meas_1, non_meas_2]

    np.testing.assert_equal(is_povm(non_meas), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test tile."""
import numpy as np

from toqito.states import basis
from toqito.states import tile


def test_tile_0():
    r"""Generate the Tile state for index = 0:
    .. math::
        |\psi_0 \rangle = \frac{1}{\sqrt{2}} |0 \rangle
        \left(|0\rangle - |1\rangle \right).
    """
    e_0, e_1 = basis(3, 0), basis(3, 1)
    expected_res = 1 / np.sqrt(2) * np.kron(e_0, (e_0 - e_1))
    res = tile(0)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tile_1():
    r"""Generate the Tile state for index = 1:
    .. math::
        |\psi_1\rangle = \frac{1}{\sqrt{2}}
        \left(|0\rangle - |1\rangle \right) |2\rangle
    """
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    expected_res = 1 / np.sqrt(2) * np.kron((e_0 - e_1), e_2)
    res = tile(1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tile_2():
    r"""Generate the Tile state for index = 2:
    .. math::
        |\psi_2\rangle = \frac{1}{\sqrt{2}} |2\rangle
        \left(|1\rangle - |2\rangle \right)
    """
    e_1, e_2 = basis(3, 1), basis(3, 2)
    expected_res = 1 / np.sqrt(2) * np.kron(e_2, (e_1 - e_2))
    res = tile(2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tile_3():
    r"""Generate the Tile state for index = 3:
    .. math::
        |\psi_3\rangle = \frac{1}{\sqrt{2}}
        \left(|1\rangle - |2\rangle \right) |0\rangle
    """
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    expected_res = 1 / np.sqrt(2) * np.kron((e_1 - e_2), e_0)
    res = tile(3)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tile_4():
    r"""Generate the Tile state for index = 4:
    .. math::
        |\psi_4\rangle = \frac{1}{3}
        \left(|0\rangle + |1\rangle + |2\rangle)\right)
        \left(|0\rangle + |1\rangle + |2\rangle.
    """
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    expected_res = 1 / 3 * np.kron((e_0 + e_1 + e_2), (e_0 + e_1 + e_2))
    res = tile(4)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tile_invalid():
    """Ensures that an integer above 4 is error-checked."""
    with np.testing.assert_raises(ValueError):
        tile(5)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Is the matrix idempotent."""
import numpy as np
from toqito.matrix_props import is_square


def is_idempotent(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-8) -> bool:
    r"""
    Check if matrix is the idempotent matrix [WikIdempotent]_.

    An *idempotent matrix* is a square matrix, which, when multiplied by itself, yields itself.
    That is, the matrix :math:`A` is idempotent if and only if :math:`A^2 = A`.

    Examples
    ==========

    The following is an example of a :math:`2 x 2` idempotent matrix:

    .. math::
        A = \begin{pmatrix}
            3 & -6 \\
            1 & -2
        \end{pmatrix}

    >>> from toqito.matrix_props import is_idempotent
    >>> import numpy as np
    >>> mat = np.array([[3, -6], [1, -2]])
    >>> is_idempotent(mat)

    Alternatively, the following matrix

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    is not idempotent.

    >>> from toqito.matrix_props import is_idempotent
    >>> import numpy as np
    >>> mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_idempotent(mat)
    False

    References
    ==========
    .. [WikIdempotent] Wikipedia: Idempotent matrix
        https://en.wikipedia.org/wiki/Idempotent_matrix

    :param mat: Matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Return :code:`True` if matrix is the idempotent matrix, and
            :code:`False` otherwise.
    """
    if not is_square(mat):
        return False
    return np.allclose(mat, mat @ mat, rtol=rtol, atol=atol)



"""Test random_ginibre."""
import numpy as np

from toqito.random import random_ginibre


def test_random_ginibre_dims():
    """Generate random Ginibre matrix and check proper dimensions."""
    gin_mat = random_ginibre(2, 2)
    np.testing.assert_equal(gin_mat.shape[0], 2)
    np.testing.assert_equal(gin_mat.shape[1], 2)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for apply_channel."""
import numpy as np

from toqito.channel_ops import apply_channel
from toqito.perms import swap_operator


def test_apply_channel_choi():
    """
    The swap operator is the Choi matrix of the transpose map.

    The following test is a (non-ideal, but illustrative) way of computing
    the transpose of a matrix.
    """
    test_input_mat = np.array([[1, 4, 7], [2, 5, 8], [3, 6, 9]])

    expected_res = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    res = apply_channel(test_input_mat, swap_operator(3))

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_apply_channel_kraus():
    """
    Apply Kraus map.

    The following test computes PHI(X) where X = [[1, 2], [3, 4]] and
    where PHI is the superoperator defined by:
    Phi(X) = [[1,5],[1,0],[0,2]] X [[0,1][2,3][4,5]].conj().T -
    [[1,0],[0,0],[0,1]] X [[0,0][1,1],[0,0]].conj().T
    """
    test_input_mat = np.array([[1, 2], [3, 4]])

    kraus_1 = np.array([[1, 5], [1, 0], [0, 2]])
    kraus_2 = np.array([[0, 1], [2, 3], [4, 5]])
    kraus_3 = np.array([[-1, 0], [0, 0], [0, -1]])
    kraus_4 = np.array([[0, 0], [1, 1], [0, 0]])

    expected_res = np.array([[22, 95, 174], [2, 8, 14], [8, 29, 64]])

    res = apply_channel(test_input_mat, [[kraus_1, kraus_2], [kraus_3, kraus_4]])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_apply_channel_invalid_input():
    """Invalid input for apply map."""
    with np.testing.assert_raises(ValueError):
        apply_channel(np.array([[1, 2], [3, 4]]), 2)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test perm_sign."""
import numpy as np

from toqito.perms import perm_sign


def test_perm_sign_small_example_even():
    """Small example when permutation is even."""
    res = perm_sign([1, 2, 3, 4])
    np.testing.assert_equal(res, 1)


def test_perm_sign_small_example_odd():
    """Small example when permutation is odd."""
    res = perm_sign([1, 2, 4, 3, 5])
    np.testing.assert_equal(res, -1)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test iden."""
import numpy as np

from toqito.matrices import iden


def test_iden_full():
    """Full 2-dimensional identity matrix."""
    expected_res = np.array([[1, 0], [0, 1]])
    res = iden(2, False)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_iden_sparse():
    """Sparse 2-dimensional identity matrix."""
    expected_res = np.array([[1, 0], [0, 1]])
    res = iden(2, True).toarray()

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Generate random unitary."""
from typing import List, Union

import numpy as np


def random_unitary(dim: Union[List[int], int], is_real: bool = False) -> np.ndarray:
    """
    Generate a random unitary or orthogonal matrix [MO09]_.

    Calculates a random unitary matrix (if :code:`is_real = False`) or a random real orthogonal
    matrix (if :code:`is_real = True`), uniformly distributed according to the Haar measure.

    Examples
    ==========

    We may generate a random unitary matrix. Here is an example of how we may be able to generate a
    random :math:`2`-dimensional random unitary matrix with complex entries.

    >>> from toqito.random import random_unitary
    >>> complex_dm = random_unitary(2)
    >>> complex_dm
    [[0.40563696+0.18092721j, 0.00066868+0.89594841j],
     [0.4237286 +0.78941628j, 0.27157521-0.35145826j]]

    We can verify that this is in fact a valid unitary matrix using the :code:`is_unitary` function
    from :code:`toqito` as follows

    >>> from toqito.matrix_props import is_unitary
    >>> is_unitary(complex_dm)
    True

    We can also generate random unitary matrices that are real-valued as follows.

    >>> from toqito.random import random_unitary
    >>> real_dm = random_unitary(2, True)
    >>> real_dm
    [[ 0.01972681, -0.99980541],
     [ 0.99980541,  0.01972681]]

    Again, verifying that this is a valid unitary matrix can be done as follows.

    >>> from toqito.matrix_props import is_unitary
    >>> is_unitary(real_dm)
    True

    We may also generate unitaries such that the dimension argument provided is a :code:`list` as
    opposed to an :code:`int`. Here is an example of a random unitary matrix of dimension :math:`4`.

    >>> from toqito.random import random_unitary
    >>> mat = random_unitary([4, 4], True)
    >>> mat
    [[ 0.48996358, -0.20978392,  0.56678587, -0.62823576],
     [ 0.62909119, -0.35852051, -0.68961425, -0.01181086],
     [ 0.38311399,  0.90865415, -0.1209574 , -0.11375677],
     [ 0.46626562, -0.04244265,  0.4342295 ,  0.76957113]]

    As before, we can verify that this matrix generated is a valid unitary matrix.

    >>> from toqito.matrix_props import is_unitary
    >>> is_unitary(mat)
    True

    References
    ==========
    .. [MO09] How to generate a random unitary matrix,
        Maris Ozols
        March 16, 2009,
        home.lu.lv/~sd20008/papers/essays/Random%20unitary%20%5Bpaper%5D.pdf

    :param dim: The number of rows (and columns) of the unitary matrix.
    :param is_real: Boolean denoting whether the returned matrix has real
                    entries or not. Default is :code:`False`.
    :return: A :code:`dim`-by-:code:`dim` random unitary matrix.
    """
    if isinstance(dim, int):
        dim = [dim, dim]

    # Construct the Ginibre ensemble.
    gin = np.random.rand(dim[0], dim[1])

    if not is_real:
        gin = gin + 1j * np.random.rand(dim[0], dim[1])

    # QR decomposition of the Ginibre ensemble.
    q_mat, r_mat = np.linalg.qr(gin)

    # Compute U from QR decomposition.
    r_mat = np.sign(np.diag(r_mat))

    # Protect against potentially zero diagonal entries.
    r_mat[r_mat == 0] = 1

    return q_mat @ np.diag(r_mat)



"""Test pure_to_mixed."""
import numpy as np

from toqito.state_ops import pure_to_mixed
from toqito.states import bell


def test_pure_to_mixed_state_vector():
    """Convert pure state to mixed state vector."""
    expected_res = np.array(
        [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
    )

    phi = bell(0)
    res = pure_to_mixed(phi)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pure_to_mixed_density_matrix():
    """Convert pure state to mixed state density matrix."""
    expected_res = np.array(
        [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
    )

    phi = bell(0) * bell(0).conj().T
    res = pure_to_mixed(phi)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_invalid_pure_to_mixed_input():
    """Invalid arguments for pure_to_mixed."""
    with np.testing.assert_raises(ValueError):
        non_valid_input = np.array([[1 / 2, 0, 0, 1 / 2], [1 / 2, 0, 0, 1 / 2]])
        pure_to_mixed(non_valid_input)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Calculate unique permutations."""
from typing import List

from dataclasses import dataclass


@dataclass
class UniqueElement:
    """Class for unique elements to keep track of occurrences."""

    value: int
    occurrences: int


def unique_perms(elements: List[int]):
    r"""
    Determine the number of unique permutations of a list.

    Examples
    ==========

    Consider the following vector

    .. math::
        \left[1, 1, 2, 2, 1, 2, 1, 3, 3, 3\right].

    The number of possible permutations possible with the above vector is :math:`4200`. This can be
    obtained using the :code:`toqito` package as follows.

    >>> from toqito.perms import unique_perms
    >>> vec_nums = [1, 1, 2, 2, 1, 2, 1, 3, 3, 3]
    >>> len(list(unique_perms(vec_nums)))
    4200

    :param elements: List of integers.
    :return: The number of possible permutations possible.
    """
    elem_set = set(elements)
    list_unique = [UniqueElement(value=i, occurrences=elements.count(i)) for i in elem_set]
    len_elems = len(elements)

    return perm_unique_helper(list_unique, [0] * len_elems, len_elems - 1)


def perm_unique_helper(list_unique: List[UniqueElement], result_list: List[int], elem_d: int):
    """
    Provide helper function for unique_perms.

    :param list_unique:
    :param result_list:
    :param elem_d:
    :return:
    """
    if elem_d < 0:
        yield tuple(result_list)
    else:
        for i in list_unique:
            if i.occurrences > 0:
                result_list[elem_d] = i.value
                i.occurrences -= 1
                for g_perm in perm_unique_helper(list_unique, result_list, elem_d - 1):
                    yield g_perm
                i.occurrences += 1



"""Test werner."""
import numpy as np

from toqito.states import werner


def test_werner_qutrit():
    """Test for qutrit Werner state."""
    res = werner(3, 1 / 2)
    np.testing.assert_equal(np.isclose(res[0][0], 0.0666666), True)
    np.testing.assert_equal(np.isclose(res[1][3], -0.066666), True)


def test_werner_multipartite():
    """Test for multipartite Werner state."""
    res = werner(2, [0.01, 0.02, 0.03, 0.04, 0.05])
    np.testing.assert_equal(np.isclose(res[0][0], 0.1127, atol=1e-02), True)


def test_werner_invalid_alpha():
    """Test for invalid `alpha` parameter."""
    with np.testing.assert_raises(ValueError):
        werner(3, [1, 2])


if __name__ == "__main__":
    np.testing.run_module_suite()



"""State purity."""
import numpy as np

from toqito.matrix_props import is_density


def purity(rho: np.ndarray) -> float:
    r"""
    Compute the purity of a quantum state [WikPurity]_.

    The negativity of a subsystem can be defined in terms of a density matrix :math:`\rho`: The
    purity of a quantum state :math:`\rho` is defined as

    .. math::
        \text{Tr}(\rho^2),

    where :math:`\text{Tr}` is the trace function.

    Examples
    ==========

    Consider the following scaled state defined as the scaled identity matrix

    .. math::
        \rho = \frac{1}{4} \begin{pmatrix}
                         1 & 0 & 0 & 0 \\
                         0 & 1 & 0 & 0 \\
                         0 & 0 & 1 & 0 \\
                         0 & 0 & 0 & 1
                       \end{pmatrix} \in \text{D}(\mathcal{X}).

    Calculating the purity of :math:`\rho` yields :math:`\frac{1}{4}`. This can be observed using
    :code:`toqito` as follows.

    >>> from toqito.state_props import purity
    >>> import numpy as np
    >>> purity(np.identity(4) / 4)
    0.25

    Calculate the purity of the Werner state.
    >>> from toqito.states import werner
    >>> rho = werner(2, 1 / 4)
    >>> purity(rho)
    0.2653

    References
    ==========
    .. [WikPurity] Wikipedia page for purity (quantum mechanics):
        https://en.wikipedia.org/wiki/Purity_(quantum_mechanics)

    :param rho: A density matrix of a pure state vector.
    :return: A value between 0 and 1 that corresponds to the purity of
            :math:`\rho`.
    """
    if not is_density(rho):
        raise ValueError("Purity is only defined for density operators.")
    # "np.real" get rid of the close-to-0 imaginary part.
    return np.real(np.trace(np.linalg.matrix_power(rho, 2)))



"""Helper functions for dealing with cvxpy objects."""
from toqito.helper.expr_as_np_array import expr_as_np_array
from toqito.helper.np_array_as_expr import np_array_as_expr
from toqito.helper.update_odometer import update_odometer
from toqito.helper.cvx_kron import cvx_kron



"""Chessboard state."""
from typing import List

import numpy as np


def chessboard(mat_params: List[float], s_param: float = None, t_param: float = None) -> np.ndarray:
    r"""
    Produce a chessboard state [BP00]_.

    Generates the chessboard state defined in [BP00]_. Note that, for certain choices of
    :code:`s_param` and :code:`t_param`, this state will not have positive partial transpose, and
    thus may not be bound entangled.

    Examples
    ==========

    The standard chessboard state can be invoked using :code:`toqito` as

    >>> from toqito.states import chessboard
    >>> chessboard([1, 2, 3, 4, 5, 6], 7, 8)
    [[ 0.22592593,  0.        ,  0.12962963,  0.        ,  0.        ,
       0.        ,  0.17777778,  0.        ,  0.        ],
     [ 0.        ,  0.01851852,  0.        ,  0.        ,  0.        ,
       0.01111111,  0.        ,  0.02962963,  0.        ],
     [ 0.12962963,  0.        ,  0.18148148,  0.        ,  0.15555556,
       0.        ,  0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.01851852,  0.        ,
       0.02222222,  0.        , -0.01481481,  0.        ],
     [ 0.        ,  0.        ,  0.15555556,  0.        ,  0.22592593,
       0.        , -0.14814815,  0.        ,  0.        ],
     [ 0.        ,  0.01111111,  0.        ,  0.02222222,  0.        ,
       0.03333333,  0.        ,  0.        ,  0.        ],
     [ 0.17777778,  0.        ,  0.        ,  0.        , -0.14814815,
       0.        ,  0.23703704,  0.        ,  0.        ],
     [ 0.        ,  0.02962963,  0.        , -0.01481481,  0.        ,
       0.        ,  0.        ,  0.05925926,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
       0.        ,  0.        ,  0.        ,  0.        ]]

    References
    ==========
    .. [BP00] Three qubits can be entangled in two inequivalent ways.
        D. Bruss and A. Peres
        Phys. Rev. A, 61:30301(R), 2000
        arXiv: 991.1056

    :param mat_params:
    :param s_param:
    :param t_param:
    :return: A chessboard state.
    """
    if s_param is None:
        s_param = np.conj(mat_params[2]) / np.conj(mat_params[5])
    if t_param is None:
        t_param = mat_params[0] * mat_params[3] / mat_params[4]

    v_1 = np.array([[mat_params[4], 0, s_param, 0, mat_params[5], 0, 0, 0, 0]])

    v_2 = np.array([[0, mat_params[0], 0, mat_params[1], 0, mat_params[2], 0, 0, 0]])

    v_3 = np.array([[np.conj(mat_params[5]), 0, 0, 0, -np.conj(mat_params[4]), 0, t_param, 0, 0]])

    v_4 = np.array(
        [
            [
                0,
                np.conj(mat_params[1]),
                0,
                -np.conj(mat_params[0]),
                0,
                0,
                0,
                mat_params[3],
                0,
            ]
        ]
    )

    rho = v_1.conj().T * v_1 + v_2.conj().T * v_2 + v_3.conj().T * v_3 + v_4.conj().T * v_4
    return rho / np.trace(rho)



"""The depolarizing channel."""
import numpy as np

from toqito.states import max_entangled


def depolarizing(dim: int, param_p: float = 0) -> np.ndarray:
    r"""
    Produce the partially depolarizing channel [WikDepo]_, [WatDepo18]_.

    The Choi matrix of the completely depolarizing channel that acts on
    :code:`dim`-by-:code:`dim` matrices.

    The *completely depolarizing channel* is defined as

    .. math::
        \Omega(X) = \text{Tr}(X) \omega

    for all :math:`X \in \text{L}(\mathcal{X})`, where

    .. math::
        \omega = \frac{\mathbb{I}_{\mathcal{X}}}{\text{dim}(\mathcal{X})}

    denotes the completely mixed stated defined with respect to the space :math:`\mathcal{X}`.

    Examples
    ==========

    The completely depolarizing channel maps every density matrix to the maximally-mixed state.
    For example, consider the density operator

    .. math::
        \rho = \frac{1}{2} \begin{pmatrix}
                             1 & 0 & 0 & 1 \\
                             0 & 0 & 0 & 0 \\
                             0 & 0 & 0 & 0 \\
                             1 & 0 & 0 & 1
                           \end{pmatrix}

    corresponding to one of the Bell states. Applying the depolarizing channel to :math:`\rho` we
    have that

    .. math::
        \Phi(\rho) = \frac{1}{4} \begin{pmatrix}
                                    \frac{1}{2} & 0 & 0 & \frac{1}{2} \\
                                    0 & 0 & 0 & 0 \\
                                    0 & 0 & 0 & 0 \\
                                    \frac{1}{2} & 0 & 0 & \frac{1}{2}
                                 \end{pmatrix}.

    This can be observed in :code:`toqito` as follows.

    >>> from toqito.channel_ops import apply_channel
    >>> from toqito.channels import depolarizing
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
    >>> )
    >>> apply_channel(test_input_mat, depolarizing(4))
    [[0.125 0.    0.    0.125]
     [0.    0.    0.    0.   ]
     [0.    0.    0.    0.   ]
     [0.125 0.    0.    0.125]]

    >>> from toqito.channel_ops import apply_channel
    >>> from toqito.channels import depolarizing
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> apply_channel(test_input_mat, depolarizing(4, 0.5))
    [[17.125  0.25   0.375  0.5  ]
     [ 0.625 17.75   0.875  1.   ]
     [ 1.125  1.25  18.375  1.5  ]
     [ 1.625  1.75   1.875 19.   ]]


    References
    ==========
    .. [WikDepo] Wikipedia: Quantum depolarizing channel
        https://en.wikipedia.org/wiki/Quantum_depolarizing

    .. [WatDepo18] Watrous, John.
        "The theory of quantum information."
        Section: "Replacement channels and the completely depolarizing channel".
        Cambridge University Press, 2018.

    :param dim: The dimensionality on which the channel acts.
    :param param_p: Default 0.
    :return: The Choi matrix of the completely depolarizing channel.
    """
    # Compute the Choi matrix of the depolarizing channel.

    # Gives a sparse non-normalized state.
    psi = max_entangled(dim=dim, is_sparse=False, is_normalized=False)
    return (1 - param_p) * np.identity(dim ** 2) / dim + param_p * (psi * psi.conj().T)



"""Distance metrics for quantum states."""
from toqito.state_metrics.trace_norm import trace_norm
from toqito.state_metrics.hilbert_schmidt import hilbert_schmidt
from toqito.state_metrics.helstrom_holevo import helstrom_holevo
from toqito.state_metrics.fidelity import fidelity
from toqito.state_metrics.sub_fidelity import sub_fidelity
from toqito.state_metrics.trace_distance import trace_distance



"""Tests for hilbert_schmidt."""
import numpy as np

from toqito.state_metrics import hilbert_schmidt
from toqito.states import bell


def test_hilbert_schmidt_bell():
    r"""Test Hilbert-Schmidt distance on two Bell states."""

    rho = bell(0) * bell(0).conj().T
    sigma = bell(3) * bell(3).conj().T

    res = hilbert_schmidt(rho, sigma)

    np.testing.assert_equal(np.isclose(res, 1), True)


def test_hilbert_schmidt_non_density_matrix():
    r"""Test Hilbert-Schmidt distance on non-density matrix."""
    rho = np.array([[1, 2], [3, 4]])
    sigma = np.array([[5, 6], [7, 8]])

    with np.testing.assert_raises(ValueError):
        hilbert_schmidt(rho, sigma)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Determine whether channel is unital."""
from typing import List, Union

import numpy as np

from toqito.channel_ops import apply_channel, kraus_to_choi
from toqito.matrix_props import is_identity


def is_unital(
    phi: Union[np.ndarray, List[List[np.ndarray]]],
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> bool:
    r"""
    Determine whether the given channel is unital [WatUnital18]_.

    A map :math:`\Phi \in \text{T} \left(\mathcal{X}, \mathcal{Y} \right)` is *unital* if it
    holds that

    .. math::
        \Phi(\mathbb{I}_{\mathcal{X}}) = \mathbb{I}_{\mathcal{Y}}.

    Examples
    ==========

    Consider the channel whose Choi matrix is the swap operator. This channel is an example of a
    unital channel.

    >>> from toqito.perms import swap_operator
    >>> from toqito.channel_props import is_unital
    >>>
    >>> choi = swap_operator(3)
    >>> is_unital(choi)
    True

    Alternatively, the channel whose Choi matrix is the depolarizing channel is an example of a
    non-unital channel.

    >>> from toqito.channels import depolarizing
    >>> from toqito.channel_props import is_unital
    >>>
    >>> choi = depolarizing(4)
    >>> is_unital(choi)
    False

    References
    ==========
    .. [WatUnital18] Watrous, John.
        "The theory of quantum information."
        Chapter: Unital channels and majorization
        Cambridge University Press, 2018.

    :param phi: The channel provided as either a Choi matrix or a list of Kraus operators.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: :code:`True` if the channel is unital, and :code:`False` otherwise.
    """
    # If the variable `phi` is provided as a list, we assume this is a list of Kraus operators.
    if isinstance(phi, list):
        phi = kraus_to_choi(phi)

    dim = int(np.sqrt(phi.shape[0]))

    # Channel is unital if :code:`mat` is the identity matrix.
    mat = apply_channel(np.identity(dim), phi)
    return is_identity(mat, rtol=rtol, atol=atol)



"""Compute the channel fidelity between two quantum channels."""
import cvxpy
import numpy as np

from toqito.channels import partial_trace


def channel_fidelity(choi_1: np.ndarray, choi_2: np.ndarray) -> float:
    r"""
    Compute the channel fidelity between two quantum channels [VW20]_.

    Let :math:`\Phi : \text{L}(\mathcal{Y}) \rightarrow \text{L}(\mathcal{X})` and
    :math:`\Psi: \text{L}(\mathcal{Y}) \rightarrow \text{L}(\mathcal{X})` be quantum channels. Then
    the root channel fidelity defined as

    .. math::
        \sqrt{F}(\Phi, \Psi) := \text{inf}_{\rho} \sqrt{F}(\Phi(\rho), \Psi(\rho))

    where :math:`\rho \in \text{D}(\mathcal{Z} \otimes \mathcal{X})` can be calculated by means of
    the following semidefinite program (Proposition 50) in [VW20]_,

    .. math::
        \begin{align*}
            \text{maximize:} \quad & \lambda \\
            \text{subject to:} \quad & \lambda \mathbb{I}_{\mathcal{Z}} \leq
                \text{Re}\left( \text{Tr}_{\mathcal{Y}} \left( Q \right) \right),\\
                & \begin{pmatrix}
                    J(\Phi) & Q^* \\
                    Q & J(\Psi)
                \end{pmatrix} \geq 0
        \end{align*}

    where :math:`Q \in \text{L}(\mathcal{Z} \otimes \mathcal{X})`.

    Examples
    ========

    For two identical channels, we should expect that the channel fidelity should yield a value of
    :math:`1`.

    >>> from toqito.channels import dephasing
    >>> from toqito.channel_metrics import channel_fidelity
    >>>
    >>> # The Choi matrices of dimension-4 for the dephasing channel
    >>> choi_1 = dephasing(4)
    >>> choi_2 = dephasing(4)
    >>> channel_fidelity(choi_1, choi_2)
    0.9999790499568767

    We can also compute the channel fidelity between two different channels. For example, we can
    compute the channel fidelity between the dephasing and depolarizing channels.

    >>> from toqito.channels import dephasing, depolarizing
    >>> from toqito.channel_metrics import channel_fidelity
    >>>
    >>> # The Choi matrices of dimension-4 for the dephasing and depolarizing channels
    >>> choi_1 = dephasing(4)
    >>> choi_2 = depolarizing(4)
    >>> channel_fidelity(choi_1, choi_2)
    0.5001368672503087

    References
    ==========
    .. [VW20] Katariya, Vishal, and Mark M. Wilde.
        "Geometric distinguishability measures limit quantum channel estimation and discrimination."
        https://arxiv.org/abs/2004.10708

    :param choi_1: The Choi matrix of the first quantum channel.
    :param choi_2: The Choi matrix of the second quantum channel.
    :return: The channel fidelity between the channels specified by the quantum channels
             corresponding to the Choi matrices :code:`choi_1` and :code:`choi_2`.
    """
    if choi_1.shape != choi_2.shape:
        raise ValueError("The Choi matrices provided should be of equal dimension.")

    choi_dim_x, choi_dim_y = choi_1.shape
    if choi_dim_x != choi_dim_y:
        raise ValueError("The Choi matrix provided must be square.")

    choi_dim = choi_dim_x
    dim = int(np.log2(choi_dim))

    lam = cvxpy.Variable(nonneg=True)
    q_var = cvxpy.Variable((choi_dim, choi_dim), complex=True)

    constraints = []
    objective = cvxpy.Maximize(lam)

    constraints.append(cvxpy.bmat([[choi_1, q_var.H], [q_var, choi_2]]) >> 0)

    constraints.append(lam * np.identity(dim) <= cvxpy.real(partial_trace(q_var, [2], [dim, dim])))

    problem = cvxpy.Problem(objective, constraints)

    return problem.solve()



"""Tests for fidelity."""
import cvxpy
import numpy as np


from toqito.state_metrics import fidelity
from toqito.states import basis


def test_fidelity_default():
    """Test fidelity default arguments."""
    rho = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
    sigma = rho

    res = fidelity(rho, sigma)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_fidelity_cvx():
    """Test fidelity for cvx objects."""
    rho = cvxpy.bmat([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
    sigma = rho

    res = fidelity(rho, sigma)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_fidelity_non_identical_states_1():
    """Test the fidelity between two non-identical states."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    sigma = 2 / 3 * e_0 * e_0.conj().T + 1 / 3 * e_1 * e_1.conj().T
    np.testing.assert_equal(np.isclose(fidelity(rho, sigma), 0.996, rtol=1e-03), True)


def test_fidelity_non_identical_states_2():
    """Test the fidelity between two non-identical states."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    sigma = 1 / 8 * e_0 * e_0.conj().T + 7 / 8 * e_1 * e_1.conj().T
    np.testing.assert_equal(np.isclose(fidelity(rho, sigma), 0.774, rtol=1e-03), True)


def test_fidelity_non_square():
    """Tests for invalid dim."""
    rho = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
    sigma = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
    with np.testing.assert_raises(ValueError):
        fidelity(rho, sigma)


def test_fidelity_invalid_dim():
    """Tests for invalid dim."""
    rho = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
    sigma = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    with np.testing.assert_raises(ValueError):
        fidelity(rho, sigma)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""A number of widely-studied quantum channels."""

from toqito.channels.choi import choi
from toqito.channels.dephasing import dephasing
from toqito.channels.depolarizing import depolarizing
from toqito.channels.partial_trace import partial_trace
from toqito.channels.partial_transpose import partial_transpose
from toqito.channels.realignment import realignment
from toqito.channels.reduction import reduction



"""Is matrix a positive definite matrix."""
import numpy as np
from toqito.matrix_props import is_hermitian


def is_positive_definite(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Check if matrix is positive definite (PD) [WikPD]_.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                2 & -1 & 0 \\
                -1 & 2 & -1 \\
                0 & -1 & 2
            \end{pmatrix}

    our function indicates that this is indeed a positive definite matrix.

    >>> from toqito.matrix_props import is_positive_definite
    >>> import numpy as np
    >>> A = np.array([[2, -1, 0], [-1, 2, -1], [0, -1, 2]])
    >>> is_positive_definite(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                -1 & -1 \\
                -1 & -1
            \end{pmatrix}

    is not positive definite.

    >>> from toqito.matrix_props import is_positive_definite
    >>> import numpy as np
    >>> B = np.array([[-1, -1], [-1, -1]])
    >>> is_positive_definite(B)
    False

    See Also
    ========
    is_positive_semidefinite

    References
    ==========
    .. [WikPD] Wikipedia: Definiteness of a matrix.
        https://en.wikipedia.org/wiki/Definiteness_of_a_matrix

    :param mat: Matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Return :code:`True` if matrix is positive definite, and :code:`False` otherwise.
    """
    if not is_hermitian(mat, rtol=rtol, atol=atol):
        return False
    evals, _ = np.linalg.eigh(mat)
    return all(x > -abs(atol) for x in evals)



"""The reduction channel."""
import numpy as np

from scipy.sparse import identity

from toqito.states import max_entangled


def reduction(dim: int, k: int = 1) -> np.ndarray:
    r"""
    Produce the reduction map or reduction channel.

    If :code:`k = 1`, this returns the Choi matrix of the reduction map which is a positive map
    on :code:`dim`-by-:code:`dim` matrices. For a different value of :code:`k`, this yields the
    Choi matrix of the map defined by:

    .. math::
        R(X) = k * \text{Tr}(X) * \mathbb{I} - X,

    where :math:`\mathbb{I}` is the identity matrix. This map is :math:`k`-positive.

    Examples
    ==========

    Using :code:`toqito`, we can generate the :math:`3`-dimensional (or standard) reduction map
    as follows.

    >>> from toqito.channels import reduction
    >>> reduction(3)
    [[ 0.,  0.,  0.,  0., -1.,  0.,  0.,  0., -1.],
     [ 0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.],
     [-1.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
     [ 0.,  0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  1.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  0.],
     [-1.,  0.,  0.,  0., -1.,  0.,  0.,  0.,  0.]])

    :param dim: A positive integer (the dimension of the reduction map).
    :param k: If this positive integer is provided, the script will instead return the Choi
              matrix of the following linear map: Phi(X) := K * Tr(X)I - X.
    :return: The reduction map.
    """
    psi = max_entangled(dim, False, False)
    return k * identity(dim ** 2) - psi * psi.conj().T



"""Sub-fidelity metric."""
import numpy as np

from toqito.matrix_props import is_density


def sub_fidelity(rho: np.ndarray, sigma: np.ndarray) -> float:
    r"""
    Compute the sub fidelity of two density matrices [MPHUZSub08]_.

    The sub-fidelity is a measure of similarity between density operators. It is defined as

    .. math::
        E(\rho, \sigma) = \text{Tr}(\rho \sigma) +
        \sqrt{2 \left[ \text{Tr}(\rho \sigma)^2 - \text{Tr}(\rho \sigma \rho \sigma) \right]},

    where :math:`\sigma` and :math:`\rho` are density matrices. The sub-fidelity serves as an lower
    bound for the fidelity.

    Examples
    ==========

    Consider the following pair of states:

    .. math::
        \rho = \frac{3}{4}|0\rangle \langle 0| +
               \frac{1}{4}|1 \rangle \langle 1|
        \sigma = \frac{1}{8}|0 \rangle \langle 0| +
                 \frac{7}{8}|1 \rangle \langle 1|.

    Calculating the fidelity between the states :math:`\rho` and :math:`\sigma` as
    :math:`F(\rho, \sigma) \approx 0.774`. This can be observed in :code:`toqito` as

    >>> from toqito.states import basis
    >>> from toqito.state_metrics import fidelity
    >>> e_0, e_1 = ket(2, 0), ket(2, 1)
    >>> rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    >>> sigma = 1/8 * e_0 * e_0.conj().T + 7/8 * e_1 * e_1.conj().T
    >>> fidelity(rho, sigma)
    0.77389339119464

    As the sub-fidelity is a lower bound on the fidelity, that is
    :math:`E(\rho, \sigma) \leq F(\rho, \sigma)`, we can use :code:`toqito` to observe that
    :math:`E(\rho, \sigma) \approx 0.599\leq F(\rho, \sigma \approx 0.774`.

    >>> from toqito.states import basis
    >>> from toqito.state_metrics import sub_fidelity
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    >>> sigma = 1/8 * e_0 * e_0.conj().T + 7/8 * e_1 * e_1.conj().T
    >>> sub_fidelity(rho, sigma)
    0.5989109809347399

    References
    ==========
    .. [MPHUZSub08] J. A. Miszczak, Z. Puchała, P. Horodecki, A. Uhlmann, K. Życzkowski
        "Sub--and super--fidelity as bounds for quantum fidelity."
        arXiv preprint arXiv:0805.2037 (2008).
        https://arxiv.org/abs/0805.2037

    :param rho: Density operator.
    :param sigma: Density operator.
    :return: The sub-fidelity between :code:`rho` and :code:`sigma`.
    """
    # Perform some error checking.
    if not np.all(rho.shape == sigma.shape):
        raise ValueError("InvalidDim: `rho` and `sigma` must be matrices of the same size.")
    if not is_density(rho) or not is_density(sigma):
        raise ValueError("Sub-fidelity is only defined for density operators.")

    return np.real(
        np.trace(rho * sigma)
        + np.sqrt(2 * (np.trace(rho * sigma) ** 2 - np.trace(rho * sigma * rho * sigma)))
    )



"""Check if states are mutually orthogonal."""
from typing import Any, List, Union

import numpy as np


def is_mutually_orthogonal(vec_list: List[Union[np.ndarray, List[Union[float, Any]]]]) -> bool:
    r"""
    Check if list of vectors are mutually orthogonal [WikOrthog]_.

    We say that two bases

    .. math::
        \begin{equation}
            \mathcal{B}_0 = \left\{u_a: a \in \Sigma \right\} \subset \mathbb{C}^{\Sigma}
            \quad \text{and} \quad
            \mathcal{B}_1 = \left\{v_a: a \in \Sigma \right\} \subset \mathbb{C}^{\Sigma}
        \end{equation}

    are *mutually orthogonal* if and only if
    :math:`\left|\langle u_a, v_b \rangle\right| = 0` for all :math:`a, b \in \Sigma`.

    For :math:`n \in \mathbb{N}`, a set of bases :math:`\left\{
    \mathcal{B}_0, \ldots, \mathcal{B}_{n-1} \right\}` are mutually orthogonal if and only if
    every basis is orthogonal with every other basis in the set, i.e. :math:`\mathcal{B}_x`
    is orthogonal with :math:`\mathcal{B}_x^{\prime}` for all :math:`x \not= x^{\prime}` with
    :math:`x, x^{\prime} \in \Sigma`.

    Examples
    ==========

    The Bell states constitute a set of mutually orthogonal vectors.

    >>> from toqito.states import bell
    >>> from toqito.state_props import is_mutually_orthogonal
    >>> states = [bell(0), bell(1), bell(2), bell(3)]
    >>> is_mutually_orthogonal(states)
    True

    The following is an example of a list of vectors that are not mutually orthogonal.

    >>> import numpy as np
    >>> from toqito.states import bell
    >>> from toqito.state_props import is_mutually_orthogonal
    >>> states = [np.array([1, 0]), np.array([1, 1])]
    >>> is_mutually_orthogonal(states)
    False

    References
    ==========
    .. [WikOrthog] Wikipedia: Orthogonality
        https://en.wikipedia.org/wiki/Orthogonality

    :param vec_list: The list of vectors to check.
    :return: True if :code:`vec_list` are mutually orthogonal, and False otherwise.
    """
    if len(vec_list) <= 1:
        raise ValueError("There must be at least two vectors provided as input.")

    for i, vec_1 in enumerate(vec_list):
        for j, vec_2 in enumerate(vec_list):
            if i != j:
                if not np.isclose(np.inner(vec_1.conj().T, vec_2.conj().T), 0):
                    return False
    return True



"""Is channel Hermiticity-preserving."""
from typing import List, Union

import numpy as np

from toqito.channel_ops import kraus_to_choi
from toqito.matrix_props import is_hermitian


def is_herm_preserving(
    phi: Union[np.ndarray, List[List[np.ndarray]]],
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> bool:
    r"""
    Determine whether the given channel is Hermitian-preserving [WatH18]_.

    A map :math:`\Phi \in \text{T} \left(\mathcal{X}, \mathcal{Y} \right)` is
    *Hermitian-preserving* if it holds that

    .. math::
        \Phi(H) \in \text{Herm}(\mathcal{Y})

    for every Hermitian operator :math:`H \in \text{Herm}(\mathcal{X})`.

    Examples
    ==========

    The map :math:`\Phi` defined as

    .. math::
        \Phi(X) = X - U X U^*

    is Hermitian-preserving, where

    .. math::
        U = \frac{1}{\sqrt{2}}
        \begin{pmatrix}
            1 & 1 \\
            -1 & 1
        \end{pmatrix}.

    >>> import numpy as np
    >>> from toqito.channel_props import is_herm_preserving
    >>> unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
    >>> kraus_ops = [[np.identity(2), np.identity(2)], [unitary_mat, -unitary_mat]]
    >>> is_herm_preserving(kraus_ops)
    True

    We may also verify whether the corresponding Choi matrix of a given map is
    Hermitian-preserving. The swap operator is the Choi matrix of the transpose map, which is
    Hermitian-preserving as can be seen as follows:

    >>> import numpy as np
    >>> from toqito.perms import swap_operator
    >>> from toqito.channel_props import is_herm_preserving
    >>> unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
    >>> choi_mat = swap_operator(3)
    >>> is_herm_preserving(choi_mat)
    True

    References
    ==========
    .. [WatH18] Watrous, John.
        "The theory of quantum information."
        Section: "Linear maps of square operators".
        Cambridge University Press, 2018.

    :param phi: The channel provided as either a Choi matrix or a list of Kraus operators.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: True if the channel is Hermitian-preserving, and False otherwise.
    """
    # If the variable `phi` is provided as a list, we assume this is a list
    # of Kraus operators.
    if isinstance(phi, list):
        phi = kraus_to_choi(phi)

    # Phi is Hermiticity-preserving if and only if its Choi matrix is Hermitian.
    if phi.shape[0] != phi.shape[1]:
        return False
    return is_hermitian(phi, rtol=rtol, atol=atol)



"""Tests for trace_norm."""
import numpy as np

from toqito.state_metrics import trace_norm
from toqito.states import basis


def test_trace_norm():
    """Test trace norm."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00 = np.kron(e_0, e_0)
    e_11 = np.kron(e_1, e_1)

    u_vec = 1 / np.sqrt(2) * (e_00 + e_11)
    rho = u_vec * u_vec.conj().T

    res = trace_norm(rho)
    _, singular_vals, _ = np.linalg.svd(rho)
    expected_res = float(np.sum(singular_vals))

    np.testing.assert_equal(np.isclose(res, expected_res), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Trace norm metric."""
import numpy as np


def trace_norm(rho: np.ndarray) -> float:
    r"""
    Compute the trace norm of the state [WikTn]_.

    The trace norm :math:`||\rho||_1` of a density matrix :math:`\rho` is the sum of the singular
    values of :math:`\rho`. The singular values are the roots of the eigenvalues of
    :math:`\rho \rho^*`.

    Examples
    ==========

    Consider the following Bell state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \in \mathcal{X}.

    The corresponding density matrix of :math:`u` may be calculated by:

    .. math::
        \rho = u u^* = \begin{pmatrix}
                         1 & 0 & 0 & 1 \\
                         0 & 0 & 0 & 0 \\
                         0 & 0 & 0 & 0 \\
                         1 & 0 & 0 & 1
                       \end{pmatrix} \in \text{D}(\mathcal{X}).

    It can be observed using :code:`toqito` that :math:`||\rho||_1 = 1` as follows.

    >>> from toqito.states import bell
    >>> from toqito.state_metrics import trace_norm
    >>> rho = bell(0) * bell(0).conj().T
    >>> trace_norm(rho)
    0.9999999999999999

    References
    ==========
    .. [WikTn] Quantiki: Trace norm
        https://www.quantiki.org/wiki/trace-norm

    :param rho: Density operator.
    :return: The trace norm of :code:`rho`.
    """
    return np.linalg.norm(rho, ord="nuc")



"""Test entanglement_of_formation."""
import numpy as np

from toqito.state_props import entanglement_of_formation
from toqito.states import basis, bell, max_mixed


def test_entanglement_of_formation_bell_state():
    """The entanglement-of-formation on a Bell state."""
    u_vec = bell(0)
    rho = u_vec * u_vec.conj().T

    res = entanglement_of_formation(rho)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_entanglement_of_formation_maximally_mixed_state():
    """The entanglement-of-formation on a maximally mixed."""
    u_vec = max_mixed(4, False)
    rho = u_vec * u_vec.conj().T

    res = entanglement_of_formation(rho)
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_entanglement_of_formation_invalid_local_dim():
    """Invalid local dimension for entanglement_of_formation."""
    with np.testing.assert_raises(ValueError):
        rho = np.identity(4)
        entanglement_of_formation(rho, 3)


def test_entanglement_of_formation_mixed_state():
    """Not presently known how to calculate for mixed states."""
    with np.testing.assert_raises(ValueError):
        e_0, e_1 = basis(2, 0), basis(2, 1)
        rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
        entanglement_of_formation(rho)


def test_entanglement_of_formation_invalid_non_square():
    """Invalid non-square matrix for entanglement_of_formation."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        entanglement_of_formation(rho)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Computes the entanglement of formation of a bipartite quantum state."""
from typing import List, Union

import numpy as np
import scipy

from toqito.channels import partial_trace
from toqito.state_props import concurrence, von_neumann_entropy


def entanglement_of_formation(rho: np.ndarray, dim: Union[List[int], int] = None) -> float:
    r"""
    Compute entanglement-of-formation of a bipartite quantum state [WikEOF]_.

    Entanglement-of-formation is the entropy of formation of the bipartite
    quantum state :code:`rho`. Note that this function currently only supports
    :code:`rho` being a pure state or a 2-qubit state: it is not known how to
    compute the entanglement-of-formation of higher-dimensional mixed states.

    This function was adapted from QETLAB.

    Examples
    ==========

    Compute the entanglement-of-formation of a Bell state.

    Let :math:`u = \frac{1}{\sqrt{2}} \left(|00\rangle + |11\rangle \right)`
    and let

    .. math::
        \rho = uu^* = \frac{1}{2}\begin{pmatrix}
                                    1 & 0 & 0 & 1 \\
                                    0 & 0 & 0 & 0 \\
                                    0 & 0 & 0 & 0 \\
                                    1 & 0 & 0 & 1
                                 \end{pmatrix}.

    The entanglement-of-formation of :math:`\rho` is equal to 1.

    >>> from toqito.state_props import entanglement_of_formation
    >>> from toqito.states import bell
    >>>
    >>> u_vec = bell(0)
    >>> rho = u_vec * u_vec.conj().T
    >>> entanglement_of_formation(rho)
    1

    References
    ==========
    .. [WikEOF] Quantiki: Entanglement-of-formation
        https://www.quantiki.org/wiki/entanglement-formation

    :param rho: A matrix or vector.
    :param dim: The default has both subsystems of equal dimension.
    :return: A value between 0 and 1 that corresponds to the
             entanglement-of-formation of :code:`\rho`.
    """
    dim_x, dim_y = rho.shape
    round_dim = int(np.round(np.sqrt(max(dim_x, dim_y))))
    eps = np.finfo(float).eps

    if dim is None:
        dim = round_dim

    # User can specify dimension as integer.
    if isinstance(dim, int):
        dim = np.array([dim, max(dim_x, dim_y) / dim], dtype=int)
        if abs(dim[1] - np.round(dim[1])) >= 2 * max(dim_x, dim_y) * eps:
            raise ValueError(
                "Invalid dimension: If `dim` is provided as a "
                "scalar, `dim` must evenly divide `len(rho)`."
            )
        dim[1] = np.round(dim[1])

    if np.prod(dim) != max(dim_x, dim_y):
        raise ValueError(
            "Invalid dimension: Please provide local dimensions " "that match the size of `rho`."
        )

    # If :code:`rho` is a rank-1 density matrix, turn it into a vector instead
    # so we can compute the entanglement-of-formation easily.
    tmp_rho = scipy.linalg.orth(rho)
    if dim_x == dim_y and tmp_rho.shape[1] == 1:
        rho = tmp_rho
        dim_y = 1

    # Start computing entanglement-of-formation.
    if min(dim_x, dim_y) == 1:
        rho = rho[:]
        return von_neumann_entropy(partial_trace(rho * rho.conj().T, 2, dim))

    # Case: :code:`rho` is a density matrix.
    if dim_x == dim_y:
        # In the two-qubit case, we know how to compute the
        # entanglement-of-formation exactly.
        if dim_x == 4:
            rho_c = concurrence(rho)

            rho_c1 = (1 + np.sqrt(1 - rho_c ** 2)) / 2
            rho_c2 = (1 - np.sqrt(1 - rho_c ** 2)) / 2

            rho_c1_log2 = 0 if rho_c1 == 0 else np.log2(rho_c1)
            rho_c2_log2 = 0 if rho_c2 == 0 else np.log2(rho_c2)

            return -rho_c1 * rho_c1_log2 - rho_c2 * rho_c2_log2
        raise ValueError(
            "Invalid dimension: It is presently only known how to compute "
            "the entanglement-of-formation for two-qubit states and pure "
            "states."
        )
    raise ValueError("Invalid dimension: `rho` must be either a vector or " "square matrix.")



"""Apply channel to an operator."""
from typing import List, Union
import numpy as np

from toqito.matrix_ops import vec
from toqito.perms import swap


def apply_channel(mat: np.ndarray, phi_op: Union[np.ndarray, List[List[np.ndarray]]]) -> np.ndarray:
    r"""
    Apply a quantum channel to an operator [WatAChan18]_.

    Specifically, an application of the channel is defined as

    .. math::
        \Phi(X) = \text{Tr}_{\mathcal{X}} \left(J(\Phi)
        \left(\mathbb{I}_{\mathcal{Y}} \otimes X^{T}\right)\right),

    where

    .. math::
        J(\Phi): \text{T}(\mathcal{X}, \mathcal{Y}) \rightarrow
        \text{L}(\mathcal{Y} \otimes \mathcal{X})

    is the Choi representation of :math:`\Phi`.

    We assume the quantum channel given as :code:`phi_op` is provided as either the Choi matrix
    of the channel or a set of Kraus operators that define the quantum channel.

    This function is adapted from the QETLAB package.

    Examples
    ==========

    The swap operator is the Choi matrix of the transpose map. The following is a (non-ideal,
    but illustrative) way of computing the transpose of a matrix.

    Consider the following matrix

    .. math::
        X = \begin{pmatrix}
                1 & 4 & 7 \\
                2 & 5 & 8 \\
                3 & 6 & 9
            \end{pmatrix}

    Applying the swap operator given as

    .. math::
        \Phi =
        \begin{pmatrix}
            1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
            0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
            0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 1
         \end{pmatrix}

    to the matrix :math:`X`, we have the resulting matrix of

    .. math::
        \Phi(X) = \begin{pmatrix}
                        1 & 2 & 3 \\
                        4 & 5 & 6 \\
                        7 & 8 & 9
                   \end{pmatrix}

    Using :code:`toqito`, we can obtain the above matrices as follows.

    >>> from toqito.channel_ops import apply_channel
    >>> from toqito.perms import swap_operator
    >>> import numpy as np
    >>> test_input_mat = np.array([[1, 4, 7], [2, 5, 8], [3, 6, 9]])
    >>> apply_channel(test_input_mat, swap_operator(3))
    [[1., 2., 3.],
     [4., 5., 6.],
     [7., 8., 9.]]

    References
    ==========
    .. [WatAChan18] Watrous, John.
        The theory of quantum information.
        Section: Representations and characterizations of channels.
        Cambridge University Press, 2018.

    :param mat: A matrix.
    :param phi_op: A superoperator. :code:`phi_op` should be provided either as a Choi matrix,
                   or as a list of numpy arrays with either 1 or 2 columns whose entries are its
                   Kraus operators.
    :return: The result of applying the superoperator :code:`phi_op` to the operator :code:`mat`.
    """
    # Both of the following methods of applying the superoperator are much faster than naively
    # looping through the Kraus operators or constructing eigenvectors of a Choi matrix.

    # The superoperator was given as a list of Kraus operators:
    if isinstance(phi_op, list):
        s_phi_op = [len(phi_op), len(phi_op[0])]

        # Map is completely positive.
        if s_phi_op[1] == 1 or (s_phi_op[0] == 1 and s_phi_op[1] > 2):
            for i in range(s_phi_op[0]):
                phi_op[i][1] = phi_op[i][0].conj().T
        else:
            for i in range(s_phi_op[0]):
                phi_op[i][1] = phi_op[i][1].conj().T
        phi_0_list = []
        phi_1_list = []
        for i in range(s_phi_op[0]):
            phi_0_list.append(phi_op[i][0])
            phi_1_list.append(phi_op[i][1])

        k_1 = np.concatenate(phi_0_list, axis=1)
        k_2 = np.concatenate(phi_1_list, axis=0)

        a_mat = np.kron(np.identity(len(phi_op)), mat)
        return k_1 @ a_mat @ k_2

    # The superoperator was given as a Choi matrix:
    if isinstance(phi_op, np.ndarray):
        mat_size = np.array(list(mat.shape))
        phi_size = np.array(list(phi_op.shape)) / mat_size

        a_mat = np.kron(vec(mat).T[0], np.identity(int(phi_size[0])))
        b_mat = np.reshape(
            swap(
                phi_op.T,
                [1, 2],
                [[mat_size[1], phi_size[1]], [mat_size[0], phi_size[0]]],
                True,
            ).T,
            (int(phi_size[0] * np.prod(mat_size)), int(phi_size[1])),
        )
        return a_mat @ b_mat
    raise ValueError(
        "Invalid: The variable `phi_op` must either be a list of "
        "Kraus operators or as a Choi matrix."
    )



"""Compute a list of Kraus operators from the Choi matrix."""
from typing import List
import numpy as np

from toqito.matrix_ops import unvec


def choi_to_kraus(choi_mat: np.ndarray, tol: float = 1e-9) -> List[List[np.ndarray]]:
    r"""
    Compute a list of Kraus operators from the Choi matrix [Rigetti20]_.

    Note that unlike the Choi or natural representation of operators, the Kraus representation is
    *not* unique.

    This function has been adapted from [Rigetti20]_.

    Examples
    ========

    Consider taking the Kraus operators of the Choi matrix that characterizes the "swap operator"
    defined as

    .. math::
        \begin{pmatrix}
        \end{pmatrix}

    The corresponding Kraus operators of the swap operator are given as follows,

    .. math::
        \begin{equation}
            \begin{aligned}
                \frac{1}{\sqrt{2}}
                \begin{pmatrix}
                    0 & i \\ -i & 0
                \end{pmatrix}, &\quad
                \frac{1}{\sqrt{2}}
                \begin{pmatrix}
                    0 & 1 \\
                    1 & 0
                \end{pmatrix}, \\
                \begin{pmatrix}
                    1 & 0 \\
                    0 & 0
                \end{pmatrix}, &\quad
                \begin{pmatrix}
                    0 & 0 \\
                    0 & 1
                \end{pmatrix}.
            \end{aligned}
        \end{equation}

    This can be verified in :code:`toqito` as follows.

    >>> import numpy as np
    >>> from toqito.channel_ops import choi_to_kraus
    >>> choi_mat = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
    >>> kraus_ops = choi_to_kraus(choi_mat)
    >>> kraus_ops
    [array([[ 0.+0.j        ,  0.+0.70710678j],
           [-0.-0.70710678j,  0.+0.j        ]]), array([[0.        , 0.70710678],
           [0.70710678, 0.        ]]), array([[1., 0.],
           [0., 0.]]), array([[0., 0.],
           [0., 1.]])]

    See Also
    ========
    kraus_to_choi

    References
    ==========
    .. [Rigetti20] Forest Benchmarking (Rigetti).
        https://github.com/rigetti/forest-benchmarking

    :param choi_mat: a dim**2 by dim**2 choi matrix
    :param tol: optional threshold parameter for eigenvalues/kraus ops to be discarded
    :return: List of Kraus operators
    """
    eigvals, v_mat = np.linalg.eigh(choi_mat)
    return [
        np.lib.scimath.sqrt(eigval) * unvec(np.array([evec]).T)
        for eigval, evec in zip(eigvals, v_mat.T)
        if abs(eigval) > tol
    ]



"""Test ppt_distinguishability."""
import numpy as np

from toqito.perms import swap_operator
from toqito.state_opt import ppt_distinguishability
from toqito.states import basis, bell


def test_ppt_distinguishability_yyd_density_matrices():
    """
    PPT distinguishing the YYD states from [1] should yield `7/8 ~ 0.875`

    Feeding the input to the function as density matrices.

    References:
    [1]: Yu, Nengkun, Runyao Duan, and Mingsheng Ying.
    "Four locally indistinguishable ququad-ququad orthogonal
    maximally entangled states."
    Physical review letters 109.2 (2012): 020506.
    https://arxiv.org/abs/1107.3224
    """
    psi_0 = bell(0)
    psi_1 = bell(2)
    psi_2 = bell(3)
    psi_3 = bell(1)

    x_1 = np.kron(psi_0, psi_0)
    x_2 = np.kron(psi_1, psi_3)
    x_3 = np.kron(psi_2, psi_3)
    x_4 = np.kron(psi_3, psi_3)

    rho_1 = x_1 * x_1.conj().T
    rho_2 = x_2 * x_2.conj().T
    rho_3 = x_3 * x_3.conj().T
    rho_4 = x_4 * x_4.conj().T

    states = [rho_1, rho_2, rho_3, rho_4]
    probs = [1 / 4, 1 / 4, 1 / 4, 1 / 4]

    # Min-error tests:
    primal_res = ppt_distinguishability(states, probs=probs, dist_method="min-error", strategy=True)
    dual_res = ppt_distinguishability(states, probs=probs, dist_method="min-error", strategy=False)

    np.testing.assert_equal(np.isclose(primal_res, 7 / 8, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, 7 / 8, atol=0.001), True)

    primal_res = ppt_distinguishability(
        states, probs=probs, dist_method="unambiguous", strategy=True
    )
    dual_res = ppt_distinguishability(
        states, probs=probs, dist_method="unambiguous", strategy=False
    )

    np.testing.assert_equal(np.isclose(primal_res, 3 / 4, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, 3 / 4, atol=0.001), True)


def test_ppt_distinguishability_yyd_vectors():
    """
    PPT distinguishing the YYD states from [1] should yield `7/8 ~ 0.875`

    Feeding the input to the function as state vectors.

    References:
    [1]: Yu, Nengkun, Runyao Duan, and Mingsheng Ying.
    "Four locally indistinguishable ququad-ququad orthogonal
    maximally entangled states."
    Physical review letters 109.2 (2012): 020506.
    https://arxiv.org/abs/1107.3224
    """
    psi_0 = bell(0)
    psi_1 = bell(2)
    psi_2 = bell(3)
    psi_3 = bell(1)

    x_1 = np.kron(psi_0, psi_0)
    x_2 = np.kron(psi_1, psi_3)
    x_3 = np.kron(psi_2, psi_3)
    x_4 = np.kron(psi_3, psi_3)

    states = [x_1, x_2, x_3, x_4]
    probs = [1 / 4, 1 / 4, 1 / 4, 1 / 4]

    primal_res = ppt_distinguishability(states, probs=probs, dist_method="min-error", strategy=True)
    dual_res = ppt_distinguishability(states, probs=probs, dist_method="min-error", strategy=False)

    np.testing.assert_equal(np.isclose(primal_res, 7 / 8, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, 7 / 8, atol=0.001), True)

    primal_res = ppt_distinguishability(
        states, probs=probs, dist_method="unambiguous", strategy=True
    )
    dual_res = ppt_distinguishability(
        states, probs=probs, dist_method="unambiguous", strategy=False
    )

    np.testing.assert_equal(np.isclose(primal_res, 3 / 4, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, 3 / 4, atol=0.001), True)


def test_ppt_distinguishability_yyd_states_no_probs():
    """
    PPT distinguishing the YYD states from [1] should yield 7/8 ~ 0.875

    If no probability vector is explicitly given, assume uniform
    probabilities are given.

    References:
    [1]: Yu, Nengkun, Runyao Duan, and Mingsheng Ying.
    "Four locally indistinguishable ququad-ququad orthogonal
    maximally entangled states."
    Physical review letters 109.2 (2012): 020506.
    https://arxiv.org/abs/1107.3224
    """
    psi_0 = bell(0)
    psi_1 = bell(2)
    psi_2 = bell(3)
    psi_3 = bell(1)

    x_1 = np.kron(psi_0, psi_0)
    x_2 = np.kron(psi_1, psi_3)
    x_3 = np.kron(psi_2, psi_3)
    x_4 = np.kron(psi_3, psi_3)

    rho_1 = x_1 * x_1.conj().T
    rho_2 = x_2 * x_2.conj().T
    rho_3 = x_3 * x_3.conj().T
    rho_4 = x_4 * x_4.conj().T

    states = [rho_1, rho_2, rho_3, rho_4]

    primal_res = ppt_distinguishability(states, probs=None, dist_method="min-error", strategy=True)
    dual_res = ppt_distinguishability(states, probs=None, dist_method="min-error", strategy=False)

    np.testing.assert_equal(np.isclose(primal_res, 7 / 8, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, 7 / 8, atol=0.001), True)

    primal_res = ppt_distinguishability(
        states, probs=None, dist_method="unambiguous", strategy=True
    )
    dual_res = ppt_distinguishability(states, probs=None, dist_method="unambiguous", strategy=False)

    np.testing.assert_equal(np.isclose(primal_res, 3 / 4, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, 3 / 4, atol=0.001), True)


def test_ppt_distinguishability_werner_hiding_pairs():
    r"""
    One quantum data hiding scheme involves the Werner hiding pair.

    A Werner hiding pair is defined by

    .. math::
    \begin{equation}
        \sigma_0^{(n)} = \frac{\mathbb{I} \otimes \mathbb{I} + W_n}{n(n+1)}
        \quad \text{and} \quad
        \sigma_1^{(n)} = \frac{\mathbb{I} \otimes \mathbb{I} - W_n}{n(n-1)}
    \end{equation}

    The optimal probability to distinguish the Werner hiding pair is known
    to be upper bounded by the following equation

    .. math::
    \begin{equation}
        \frac{1}{2} + \frac{1}{n+1}
    \end{equation}

    References:
    [1]: Terhal, Barbara M., David P. DiVincenzo, and Debbie W. Leung.
    "Hiding bits in Bell states."
    Physical review letters 86.25 (2001): 5807.
    https://arxiv.org/abs/quant-ph/0011042

    [2]: Cosentino, Alessandro
    "Quantum state local distinguishability via convex optimization".
    University of Waterloo, Thesis
    https://uwspace.uwaterloo.ca/handle/10012/9572
    """
    dim = 2
    sigma_0 = (np.kron(np.identity(dim), np.identity(dim)) + swap_operator(dim)) / (dim * (dim + 1))
    sigma_1 = (np.kron(np.identity(dim), np.identity(dim)) - swap_operator(dim)) / (dim * (dim - 1))

    states = [sigma_0, sigma_1]

    expected_val = 1 / 2 + 1 / (dim + 1)

    primal_res = ppt_distinguishability(states, probs=None, dist_method="min-error", strategy=True)
    dual_res = ppt_distinguishability(states, probs=None, dist_method="min-error", strategy=False)

    np.testing.assert_equal(np.isclose(primal_res, expected_val, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, expected_val, atol=0.001), True)

    primal_res = ppt_distinguishability(
        states, probs=None, dist_method="unambiguous", strategy=True
    )
    dual_res = ppt_distinguishability(states, probs=None, dist_method="unambiguous", strategy=False)

    np.testing.assert_equal(np.isclose(primal_res, 1 / 3, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, 1 / 3, atol=0.001), True)


def test_ppt_distinguishability_four_bell_states():
    r"""
    PPT distinguishing the four Bell states.

    There exists a closed form formula for the probability with which one
    is able to distinguish one of the four Bell states given with equal
    probability when Alice and Bob have access to a resource state [1].

    The resource state is defined by

    ..math::
        |\tau_{\epsilon} \rangle = \sqrt{\frac{1+\epsilon}{2}} +
        |0\rangle | 0\rangle +
        \sqrt{\frac{1-\epsilon}{2}} |1 \rangle |1 \rangle

    The closed form probability with which Alice and Bob can distinguish via
    PPT measurements is given as follows

    .. math::
        \frac{1}{2} \left(1 + \sqrt{1 - \epsilon^2} \right).

    This formula happens to be equal to LOCC and SEP as well for this case.
    Refer to Theorem 5 in [1] for more details.

    References:
    [1]: Bandyopadhyay, Somshubhro, et al.
    "Limitations on separable measurements by convex optimization."
    IEEE Transactions on Information Theory 61.6 (2015): 3593-3604.
    https://arxiv.org/abs/1408.6981
    """
    rho_1 = bell(0) * bell(0).conj().T
    rho_2 = bell(1) * bell(1).conj().T
    rho_3 = bell(2) * bell(2).conj().T
    rho_4 = bell(3) * bell(3).conj().T

    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00 = np.kron(e_0, e_0)
    e_11 = np.kron(e_1, e_1)

    eps = 0.5
    resource_state = np.sqrt((1 + eps) / 2) * e_00 + np.sqrt((1 - eps) / 2) * e_11
    resource_state = resource_state * resource_state.conj().T

    states = [
        np.kron(rho_1, resource_state),
        np.kron(rho_2, resource_state),
        np.kron(rho_3, resource_state),
        np.kron(rho_4, resource_state),
    ]
    probs = [1 / 4, 1 / 4, 1 / 4, 1 / 4]

    exp_res = 1 / 2 * (1 + np.sqrt(1 - eps ** 2))

    primal_res = ppt_distinguishability(states, probs=probs, dist_method="min-error", strategy=True)
    dual_res = ppt_distinguishability(states, probs=probs, dist_method="min-error", strategy=False)

    np.testing.assert_equal(np.isclose(primal_res, exp_res, atol=0.001), True)
    np.testing.assert_equal(np.isclose(dual_res, exp_res, atol=0.001), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test random_density_matrix."""
import numpy as np

from toqito.random import random_density_matrix
from toqito.matrix_props import is_density


def test_random_density_not_real():
    """Generate random non-real density matrix."""
    mat = random_density_matrix(2)
    np.testing.assert_equal(is_density(mat), True)


def test_random_density_real():
    """Generate random real density matrix."""
    mat = random_density_matrix(2, True)
    np.testing.assert_equal(is_density(mat), True)


def test_random_density_not_real_bures():
    """Random non-real density matrix according to Bures metric."""
    mat = random_density_matrix(2, distance_metric="bures")
    np.testing.assert_equal(is_density(mat), True)


def test_random_density_not_real_k_param():
    """Generate random non-real density matrix wih k_param."""
    mat = random_density_matrix(2, distance_metric="bures")
    np.testing.assert_equal(is_density(mat), True)


def test_random_density_not_real_all_params():
    """Generate random non-real density matrix all params."""
    mat = random_density_matrix(2, True, 2, "haar")
    np.testing.assert_equal(is_density(mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""PPT symmetric extension hierarchy."""
from typing import List

import cvxpy
import numpy as np
from toqito.channels import partial_trace, partial_transpose
from toqito.perms import symmetric_projection
from .state_helper import __is_states_valid, __is_probs_valid


def symmetric_extension_hierarchy(
    states: List[np.ndarray], probs: List[float] = None, level: int = 2
) -> float:
    r"""
    Compute optimal value of the symmetric extension hierarchy SDP [Nav08]_.

    The probability of distinguishing a given set of states via PPT measurements serves as a natural
    upper bound to the value of obtaining via separable measurements. Due to the nature of separable
    measurements, it is not possible to optimize directly over these objects via semidefinite
    programming techniques.

    We can, however, construct a hierarchy of semidefinite programs that attains closer and closer
    approximations at the separable value via the techniques described in [Nav08].

    The mathematical form of this hierarchy implemented here is explicitly given from equation 4.55
    in [Cos15]_.

    .. math::

        \begin{equation}
            \begin{aligned}
                \text{maximize:} \quad & \sum_{k=1}^N p_k \langle \rho_k, \mu(k) \rangle, \\
                \text{subject to:} \quad & \sum_{k=1}^N \mu(k) =
                                           \mathbb{I}_{\mathcal{X} \otimes \mathcal{Y}}, \\
                                        & \text{Tr}_{\mathcal{Y}_2 \otimes \ldots \otimes
                                          \mathcal{Y}_s}(X_k) = \mu(k), \\
                                        & \left( \mathbb{I}_{\mathcal{X}} \otimes
                                          \Pi_{\mathcal{Y} \ovee \mathcal{Y}_2 \ovee \ldots \ovee
                                          \mathcal{Y}_s} \right) X_k
                                          \left(\mathbb{I}_{\mathcal{X}} \otimes
                                          \Pi_{\mathcal{Y} \ovee \mathcal{Y}_2 \ovee \ldots \ovee
                                          \mathcal{Y}_s} \right)
                                          = X_k \\
                                        & \text{T}_{\mathcal{X}}(X_k) \in \text{Pos}\left(
                                            \mathcal{X} \otimes \mathcal{Y} \otimes \mathcal{Y}_2
                                            \otimes \ldots \otimes \mathcal{Y}_s \right), \\
                                        & \text{T}_{\mathcal{Y}_2 \otimes \ldots \otimes
                                            \mathcal{Y}_s}(X_k) \in \text{Pos}\left(
                                            \mathcal{X} \otimes \mathcal{Y} \otimes \mathcal{Y}_2
                                            \otimes \ldots \otimes \mathcal{Y}_s \right), \\
                                        & X_1, \ldots, X_N \in
                                          \text{Pos}\left(\mathcal{X} \otimes \mathcal{Y} \otimes
                                          \mathcal{Y}_2 \otimes \ldots \otimes \mathcal{Y}_s
                                          \right).
            \end{aligned}
        \end{equation}

    Examples
    ==========

    It is known from [Cos15]_ that distinguishing three Bell states along with a resource state
    :math:`|\tau_{\epsilon}\rangle` via separable measurements has the following closed form

    .. math::
        \frac{1}{3} \left(2 + \sqrt{1 - \epsilon^2} \right)

    where the resource state is defined as

    .. math::
        |\tau_{\epsilon} \rangle = \sqrt{\frac{1+\epsilon}{2}} |00\rangle +
                                   \sqrt{\frac{1-\epsilon}{2}} |11\rangle.

    The value of optimally distinguishing these states via PPT measurements is strictly larger than
    the value one obtains from separable measurements. Calculating the first level of the hierarchy
    provides for us the optimal value of PPT measurements.

    Consider a fixed value of :math:`\epsilon = 0.5`.

    >>> from toqito.states import basis, bell
    >>> from toqito.perms import swap
    >>> import numpy as np
    >>>
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> e_00, e_11 = np.kron(e_0, e_0), np.kron(e_1, e_1)
    >>>
    >>> # Define the resource state.
    >>> eps = 0.5
    >>> eps_state = np.sqrt((1+eps)/2) * e_00 + np.sqrt((1-eps)/2) * e_11
    >>> eps_dm = eps_state * eps_state.conj().T
    >>>
    >>> # Define the ensemble of states to be distinguished.
    >>> states = [
    >>>     np.kron(bell(0) * bell(0).conj().T, eps_dm),
    >>>     np.kron(bell(1) * bell(1).conj().T, eps_dm),
    >>>     np.kron(bell(2) * bell(2).conj().T, eps_dm),
    >>>     np.kron(bell(3) * bell(3).conj().T, eps_dm),
    >>> ]
    >>>
    >>> # Ensure the distinguishability is conducted on the proper spaces.
    >>> states = [
    >>>     swap(states[0], [2, 3], [2, 2, 2, 2])
    >>>     swap(states[1], [2, 3], [2, 2, 2, 2])
    >>>     swap(states[2], [2, 3], [2, 2, 2, 2])
    >>> ]
    >>>
    >>> # Calculate the first level of the symmetric extension hierarchy. This
    >>> # is simply the value of optimally distinguishing via PPT measurements.
    >>> symmetric_extension_hierarchy(states=states, probs=None, level=1)
    0.9915817434994775
    >>>
    >>> # Calculating the second value gets closer to the separable value.
    >>> symmetric_extension_hierarchy(states=states, probs=None, level=2)
    0.958305796189204
    >>>
    >>> # As proven in [Cos15]_, the true separable value of distinguishing the
    >>> # three Bell states is:
    >>> 1/3 * (2 + np.sqrt(1 - eps**2))
    0.9553418012614794
    >>>
    >>> # Computing further levels of the hierarchy would eventually converge to
    >>> # this value, however, the higher the level, the more computationally
    >>> # demanding the SDP becomes.

    References
    ==========
    .. [Nav08] Navascués, Miguel.
        "Pure state estimation and the characterization of entanglement."
        Physical review letters 100.7 (2008): 070503.
        https://arxiv.org/abs/0707.4398

    .. [Cos15] Cosentino, Alessandro.
        "Quantum State Local Distinguishability via Convex Optimization"
        The University of Waterloo, Ph.D. Dissertation, 2015.
        https://uwspace.uwaterloo.ca/handle/10012/9572

    :param states: A list of states provided as either matrices or vectors.
    :param probs: Respective list of probabilities each state is selected.
    :param level: Level of the hierarchy to compute.
    :return: The optimal probability of the symmetric extension hierarchy SDP for level
            :code:`level`.
    """
    obj_func = []
    meas = []
    x_var = []
    constraints = []

    __is_states_valid(states)
    if probs is None:
        probs = [1 / len(states)] * len(states)
    __is_probs_valid(probs)

    dim_x, dim_y = states[0].shape

    # The variable `states` is provided as a list of vectors. Transform them
    # into density matrices.
    if dim_y == 1:
        for i, state_ket in enumerate(states):
            states[i] = state_ket * state_ket.conj().T

    dim = int(np.log2(dim_x))
    dim_list = [dim] * (level + 1)
    # The `sys_list` variable contains the numbering pertaining to the symmetrically extended
    # spaces.
    sys_list = list(range(3, 3 + level - 1))
    sym = symmetric_projection(dim, level)

    dim_xy = dim_x
    dim_xyy = np.prod(dim_list)
    for k, _ in enumerate(states):
        meas.append(cvxpy.Variable((dim_xy, dim_xy), PSD=True))
        x_var.append(cvxpy.Variable((dim_xyy, dim_xyy), PSD=True))
        constraints.append(partial_trace(x_var[k], sys_list, dim_list) == meas[k])
        constraints.append(
            np.kron(np.identity(dim), sym) @ x_var[k] @ np.kron(np.identity(dim), sym) == x_var[k]
        )
        constraints.append(partial_transpose(x_var[k], 1, dim_list) >> 0)
        for sys in range(level - 1):
            constraints.append(partial_transpose(x_var[k], sys + 3, dim_list) >> 0)

        obj_func.append(probs[k] * cvxpy.trace(states[k].conj().T @ meas[k]))

    constraints.append(sum(meas) == np.identity(dim_xy))

    objective = cvxpy.Maximize(sum(obj_func))
    problem = cvxpy.Problem(objective, constraints)
    sol_default = problem.solve()

    return sol_default



"""Test chessboard."""
import numpy as np

from toqito.states import chessboard


def test_chessboard():
    """The chessboard_state."""
    res = chessboard([1, 2, 3, 4, 5, 6], 7, 8)
    np.testing.assert_equal(np.isclose(res[0][0], 0.22592592592592592), True)


def test_chessboard_default_s():
    """The chessboard_state with default `s_param`."""
    res = chessboard([1, 2, 3, 4, 5, 6], 7)
    np.testing.assert_equal(np.isclose(res[0][0], 0.29519938056523426), True)


def test_chessboard_default_s_t():
    """The chessboard_state with default `s_param` and `t_param`."""
    res = chessboard([1, 2, 3, 4, 5, 6])
    np.testing.assert_equal(np.isclose(res[0][0], 0.3863449236810438), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Werner state."""
from typing import List, Union

import itertools
import numpy as np

from toqito.perms import permutation_operator
from toqito.perms import swap_operator


def werner(dim: int, alpha: Union[float, List[float]]) -> np.ndarray:
    r"""
    Produce a Werner state [Wer89]_.

    A Werner state is a state of the following form

    .. math::

        \begin{equation}
            \rho_{\alpha} = \frac{1}{d^2 - d\alpha} \left(\mathbb{I} \otimes
            \mathbb{I} - \alpha S \right) \in \mathbb{C}^d \otimes \mathbb{C}^d.
        \end{equation}

    Yields a Werner state with parameter :code:`alpha` acting on :code:`(dim * dim)`- dimensional
    space. More specifically, :math:`\rho` is the density operator defined by
    :math:`(\mathbb{I} - `alpha` S)` (normalized to have trace 1), where :math:`\mathbb{I}` is the
    density operator and :math:`S` is the operator that swaps two copies of :code:`dim`-dimensional
    space (see swap and swap_operator for example).

    If :code:`alpha` is a vector with :math:`p!-1` entries, for some integer :math:`p > 1`, then a
    multipartite Werner state is returned. This multipartite Werner state is the normalization of
    I - `alpha(1)*P(2)` - ... - `alpha(p!-1)*P(p!)`, where P(i) is the operator that permutes p
    subsystems according to the i-th permutation when they are written in lexicographical order
    (for example, the lexicographical ordering when p = 3 is:
    `[1, 2, 3], [1, 3, 2], [2, 1,3], [2, 3, 1], [3, 1, 2], [3, 2, 1],`

    so P(4) in this case equals permutation_operator(dim, [2, 3, 1]).

    Examples
    ==========

    Computing the qutrit Werner state with :math:`\alpha = 1/2` can be done in :code:`toqito` as

    >>> from toqito.states import werner
    >>> werner(3, 1 / 2)
    [[ 0.06666667,  0.        ,  0.        ,  0.        ,  0.        ,
       0.        ,  0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.13333333,  0.        , -0.06666667,  0.        ,
       0.        ,  0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.13333333,  0.        ,  0.        ,
       0.        , -0.06666667,  0.        ,  0.        ],
     [ 0.        , -0.06666667,  0.        ,  0.13333333,  0.        ,
       0.        ,  0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.06666667,
       0.        ,  0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
       0.13333333,  0.        , -0.06666667,  0.        ],
     [ 0.        ,  0.        , -0.06666667,  0.        ,  0.        ,
       0.        ,  0.13333333,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
       -0.06666667,  0.        ,  0.13333333,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
       0.        ,  0.        ,  0.        ,  0.06666667]]

    We may also compute multipartite Werner states in :code:`toqito` as well.

    >>> from toqito.states import werner
    >>> werner(2, [0.01, 0.02, 0.03, 0.04, 0.05])
    [[ 0.12179487,  0.        ,  0.        ,  0.        ,  0.        ,
       0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.12820513,  0.        ,  0.        , -0.00641026,
       0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.12179487,  0.        ,  0.        ,
       0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.12820513,  0.        ,
       0.        , -0.00641026,  0.        ],
     [ 0.        , -0.00641026,  0.        ,  0.        ,  0.12820513,
       0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
       0.12179487,  0.        ,  0.        ],
     [ 0.        ,  0.        ,  0.        , -0.00641026,  0.        ,
       0.        ,  0.12820513,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
       0.        ,  0.        ,  0.12179487]]

    References
    ==========
    .. [Wer89] R. F. Werner.
        Quantum states with Einstein-Podolsky-Rosen correlations admitting a
        hidden-variable model. Phys. Rev. A, 40(8):4277–4281. 1989

    :param dim: The dimension of the Werner state.
    :param alpha: Parameter to specify Werner state.
    :return: A Werner state of dimension :code:`dim`.
    """
    # The total number of permutation operators.
    if isinstance(alpha, float):
        n_fac = 2
    else:
        n_fac = len(alpha) + 1

    # Multipartite Werner state.
    if n_fac > 2:
        # Compute the number of parties from `len(alpha)`.
        n_var = n_fac
        # We won't actually go all the way to `n_fac`.
        for i in range(2, n_fac):
            n_var = n_var // i
            if n_var == i + 1:
                break
            if n_var < i:
                raise ValueError(
                    "InvalidAlpha: The `alpha` vector must contain"
                    " p!-1 entries for some integer p > 1."
                )

        # Done error checking and computing the number of parties -- now
        # compute the Werner state.
        perms = list(itertools.permutations(np.arange(n_var)))
        sorted_perms = np.argsort(perms, axis=1) + 1

        for i in range(2, n_fac):
            rho = np.identity(dim ** n_var) - alpha[i - 1] * permutation_operator(
                dim, sorted_perms[i, :], False, True
            )
        rho = rho / np.trace(rho)
        return rho
    # Bipartite Werner state.
    return (np.identity(dim ** 2) - alpha * swap_operator(dim, True)) / (dim * (dim - alpha))



"""Check if violates the PPT criterion."""
from typing import List, Union

import numpy as np

from toqito.matrix_props import is_positive_semidefinite
from toqito.channels import partial_transpose


def is_ppt(
    mat: np.ndarray, sys: int = 2, dim: Union[int, List[int]] = None, tol: float = None
) -> bool:
    r"""
    Determine whether or not a matrix has positive partial transpose [WikPPT]_.

    Yields either :code:`True` or :code:`False`, indicating that :code:`mat` does or does not have
    positive partial transpose (within numerical error). The variable :code:`mat` is assumed to act
    on bipartite space.

    For shared systems of :math:`2 \otimes 2` or :math:`2 \otimes 3`, the PPT criterion serves as a
    method to determine whether a given state is entangled or separable. Therefore, for systems of
    this size, the return value :code:`True` would indicate that the state is separable and a value
    of :code:`False` would indicate the state is entangled.

    Examples
    ==========

    Consider the following matrix

    .. math::
        X =
        \begin{pmatrix}
            1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 \\
        \end{pmatrix}.

    This matrix trivially satisfies the PPT criterion as can be seen using the
    :code:`toqito` package.

    >>> from toqito.state_props import is_ppt
    >>> import numpy as np
    >>> mat = np.identity(9)
    >>> is_ppt(mat)
    True

    Consider the following Bell state:

    .. math::
        u = \frac{1}{\sqrt{2}}\left( |01 \rangle + |10 \rangle \right).

    For the density matrix :math:`\rho = u u^*`, as this is an entangled state
    of dimension :math:`2`, it will violate the PPT criterion, which can be seen
    using the :code:`toqito` package.

    >>> from toqito.states import bell
    >>> from toqito.state_props import is_ppt
    >>> rho = bell(2) * bell(2).conj().T
    >>> is_ppt(rho)
    False

    References
    ==========
    .. [WikPPT] Quantiki: Positive partial transpose
        https://www.quantiki.org/wiki/positive-partial-transpose

    :param mat: A square matrix.
    :param sys: Scalar or vector indicating which subsystems the transpose
                should be applied on.
    :param dim: The dimension is a vector containing the dimensions of the
                subsystems on which :code:`mat` acts.
    :param tol: Tolerance with which to check whether `mat` is PPT.
    :return: Returns :code:`True` if :code:`mat` is PPT and :code:`False` if
             not.
    """
    eps = np.finfo(float).eps

    sqrt_rho_dims = np.round(np.sqrt(list(mat.shape)))

    if dim is None:
        dim = np.array([[sqrt_rho_dims[0], sqrt_rho_dims[0]], [sqrt_rho_dims[1], sqrt_rho_dims[1]]])
    if tol is None:
        tol = np.sqrt(eps)
    return is_positive_semidefinite(partial_transpose(mat, sys, dim), tol)



"""Tests for is_herm_preserving."""
import numpy as np

from toqito.channel_props import is_herm_preserving
from toqito.perms import swap_operator


def test_is_herm_preserving_kraus_true():
    """Verify Hermitian-preserving channel as Kraus ops as True."""
    unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
    kraus_ops = [[np.identity(2), np.identity(2)], [unitary_mat, -unitary_mat]]

    np.testing.assert_equal(is_herm_preserving(kraus_ops), True)


def test_is_herm_preserving_kraus_false():
    """Verify non-Hermitian-preserving channel as Kraus ops as False."""
    unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
    kraus_ops = [[np.identity(2), unitary_mat], [unitary_mat, -unitary_mat]]

    np.testing.assert_equal(is_herm_preserving(kraus_ops), False)


def test_is_herm_preserving_choi_true():
    """Swap operator is Choi matrix of the (Herm-preserving) transpose map."""
    choi_mat = swap_operator(3)
    np.testing.assert_equal(is_herm_preserving(choi_mat), True)


def test_is_herm_preserving_non_square():
    """Verify non-square Choi matrix returns False."""
    non_square_mat = np.array([[1, 2, 3], [4, 5, 6]])
    np.testing.assert_equal(is_herm_preserving(non_square_mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Determine probability of obtaining measurement outcome."""
import numpy as np


def measure(measurement: np.ndarray, state: np.ndarray) -> float:
    r"""
    Determine probability of obtaining a measurement outcome applied to state.

    A *measurement* is defined by a function

    .. math::
        \mu : \Sigma \rightarrow \text{Pos}(\mathcal{X}),

    for some choice of alphabet :math:`\Sigma` and a complex Euclidean space :math:`\mathcal{X}`
    that satisfies

    .. math::
        \sum_{a \in \Sigma} \mu(a) = \mathbb{I}_{\mathcal{X}}.

    Further information can be found here [WikMeas]_.

    Examples
    ==========

    Consider the following state:

    .. math::
        u = \frac{1}{\sqrt{3}} e_0 + \sqrt{\frac{2}{3}} e_1

    where we define :math:`u u^* = \rho \in \text{D}(\mathcal{X})`.

    Define measurement operators

    .. math::
        P_0 = e_0 e_0^* \quad \text{and} \quad P_1 = e_1 e_1^*.

    >>> from toqito.states import basis
    >>> from toqito.measurement_ops import measure
    >>> import numpy as np
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>>
    >>> u = 1/np.sqrt(3) * e_0 + np.sqrt(2/3) * e_1
    >>> rho = u * u.conj().T
    >>>
    >>> proj_0 = e_0 * e_0.conj().T
    >>> proj_1 = e_1 * e_1.conj().T

    Then the probability of obtaining outcome :math:`0` is given by

    .. math::
        \langle P_0, \rho \rangle = \frac{1}{3}.

    >>> measure(proj_0, rho)
    0.3333333333333334

    Similarly, the probability of obtaining outcome math:`1` is given by

    .. math::
        \langle P_1, \rho \rangle = \frac{2}{3}.

    >>> measure(proj_1, rho)
    0.6666666666666666

    References
    ==========
    .. [WikMeas] Wikipedia: Measurement in quantum mechanics
        https://en.wikipedia.org/wiki/Measurement_in_quantum_mechanics

    :param measurement: The measurement to apply.
    :param state: The state to apply the measurement to.
    :return: Returns the probability of obtaining a given outcome after applying
             the variable :code:`measurement` to the variable :code:`state`.
    """
    return float(np.trace(measurement.conj().T * state))



"""Check if state is product vector."""
from typing import List, Union

import numpy as np

from toqito.state_ops import schmidt_decomposition


def is_product_vector(vec: np.ndarray, dim: Union[int, List[int]] = None) -> bool:
    r"""
    Determine if a given vector is a product vector [WikProdState]_.

    Examples
    ==========

    Consider the following Bell state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \in \mathcal{X}.

    The corresponding density matrix of :math:`u` may be calculated by:

    .. math::
        \rho = u u^* = \frac{1}{2} \begin{pmatrix}
                         1 & 0 & 0 & 1 \\
                         0 & 0 & 0 & 0 \\
                         0 & 0 & 0 & 0 \\
                         1 & 0 & 0 & 1
                       \end{pmatrix} \in \text{D}(\mathcal{X}).

    Calculating the rank of :math:`\rho` yields that the :math:`\rho` is a pure state. This can be
    confirmed in :code:`toqito` as follows:

    >>> from toqito.states import bell
    >>> from toqito.state_props import is_pure
    >>> u = bell(0)
    >>> rho = u * u.conj().T
    >>> is_pure(rho)
    True

    It is also possible to determine whether a set of density matrices are pure. For instance, we
    can see that the density matrices corresponding to the four Bell states yield a result of
    :code:`True` indicating that all states provided to the function are pure.

    >>> from toqito.states import bell
    >>> from toqito.state_props import is_pure
    >>> u0, u1, u2, u3 = bell(0), bell(1), bell(2), bell(3)
    >>> rho0 = u0 * u0.conj().T
    >>> rho1 = u1 * u1.conj().T
    >>> rho2 = u2 * u2.conj().T
    >>> rho3 = u3 * u3.conj().T
    >>>
    >>> is_pure([rho0, rho1, rho2, rho3])
    True

    References
    ==========
    .. [WikProdState] Wikipedia: Quantum state - Pure states
        https://en.wikipedia.org/wiki/Quantum_state#Pure_states

    :param vec: The vector to check.
    :param dim: The dimension of the vector
    :return: True if :code:`vec` is a product vector and False otherwise.
    """
    return _is_product_vector(vec, dim)[0][0]


def _is_product_vector(vec: np.ndarray, dim: Union[int, List[int]] = None) -> [int, bool]:
    """
    Determine if a given vector is a product vector recursive helper.

    :param vec: The vector to check.
    :param dim: The dimension of the vector
    :return: :code:`True` if :code:`vec` is a product vector and :code:`False`
             otherwise.
    """
    if dim is None:
        dim = np.round(np.sqrt(len(vec)))
    if isinstance(dim, list):
        dim = np.array(dim)

    # Allow the user to enter a single number for dim.
    if isinstance(dim, float):
        num_sys = 1
    else:
        num_sys = len(dim)

    if num_sys == 1:
        dim = np.array([dim, len(vec) // dim])
        dim[1] = np.round(dim[1])
        num_sys = 2

    dec = 0
    # If there are only two subsystems, just use the Schmidt decomposition.
    if num_sys == 2:
        singular_vals, u_mat, vt_mat = schmidt_decomposition(vec, dim, 2)
        ipv = singular_vals[1] <= np.prod(dim) * np.spacing(singular_vals[0])

        # Provide this even if not requested, since it is needed if this
        # function was called as part of its recursive algorithm (see below)
        if ipv:
            u_mat = u_mat * np.sqrt(singular_vals[0])
            vt_mat = vt_mat * np.sqrt(singular_vals[0])
            dec = [u_mat[:, 0], vt_mat[:, 0]]
    else:
        new_dim = [dim[0] * dim[1]]
        new_dim.extend(dim[2:])
        ipv, dec = _is_product_vector(vec, new_dim)
        if ipv:
            ipv, tdec = _is_product_vector(dec[0], [dim[0], dim[1]])
            if ipv:
                dec = [tdec, dec[1:]]

    return ipv, dec



"""Test is_mutually_orthogonal."""
import numpy as np

from toqito.states import bell
from toqito.state_props import is_mutually_orthogonal


def test_is_mutually_orthogonal_bell_states():
    """Return True for orthogonal Bell vectors."""
    states = [bell(0), bell(1), bell(2), bell(3)]
    np.testing.assert_equal(is_mutually_orthogonal(states), True)


def test_is_not_mutually_orthogonal():
    """Return False for non-orthogonal vectors."""
    states = [np.array([1, 0]), np.array([1, 1])]
    np.testing.assert_equal(is_mutually_orthogonal(states), False)


def test_is_mutually_orthogonal_basis_invalid_input_len():
    """Tests for invalid input len."""
    with np.testing.assert_raises(ValueError):
        vec_list = [np.array([1, 0])]
        is_mutually_orthogonal(vec_list)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Check if set of states form an ensemble."""
from typing import List

import numpy as np

from toqito.matrix_props import is_positive_semidefinite


def is_ensemble(states: List[np.ndarray]) -> bool:
    r"""
    Determine if a set of states constitute an ensemble [WatEns18]_.

    An ensemble of quantum states is defined by a function

    .. math::
        \eta : \Gamma \rightarrow \text{Pos}(\mathcal{X})

    that satisfies

    .. math::
        \text{Tr}\left( \sum_{a \in \Gamma} \eta(a) \right) = 1.

    Examples
    ==========

    Consider the following set of matrices

    .. math::
        \eta = \left\{ \rho_0, \rho_1 \right\}

    where

    .. math::
        \rho_0 = \frac{1}{2} \begin{pmatrix} 1 & 0 \\ 0 & 0 \end{pmatrix}, \quad
        \rho_1 = \frac{1}{2} \begin{pmatrix} 0 & 0 \\ 0 & 1 \end{pmatrix}.

    The set :math:`\eta` constitutes a valid ensemble.

    >>> from toqito.state_props import is_ensemble
    >>> import numpy as np
    >>> rho_0 = np.array([[0.5, 0], [0, 0]])
    >>> rho_1 = np.array([[0, 0], [0, 0.5]])
    >>> states = [rho_0, rho_1]
    >>> is_ensemble(states)
    True

    References
    ==========
    .. [WatEns18] Watrous, John.
        "The theory of quantum information."
        Section: "Ensemble of quantum states".
        Cambridge University Press, 2018.

    :param states: The list of states to check.
    :return: True if states form an ensemble and False otherwise.
    """
    trace_sum = 0
    for state in states:
        trace_sum += np.trace(state)
        # Constraint: All states in ensemble must be positive semidefinite.
        if not is_positive_semidefinite(state):
            return False
    # Constraint: The sum of the traces of all states within the ensemble must
    # be equal to 1.
    return np.allclose(trace_sum, 1)
